"""
=========================
Data Validation Utilities
=========================

This module contains utility functions for validating datasets, e.g. the validation
function(s) for processed data being passed out of one pipeline step and into the next.

"""

from pathlib import Path

import pandas as pd
from pandas.api.types import is_integer_dtype
from pyarrow import parquet as pq


def _read_file(filepath: str) -> pd.DataFrame:
    """Reads a file.

    Parameters
    ----------
    filepath : str
        The path to the file to read.

    Returns
    -------
        The loaded DataFrame.

    Raises
    ------
    NotImplementedError
        If the file type is not supported.
    """
    extension = Path(filepath).suffix
    if extension == ".parquet":
        return pd.read_parquet(filepath)
    elif extension == ".csv":
        return pd.read_csv(filepath)
    else:
        raise NotImplementedError(
            f"Data file type {extension} is not supported. Convert to Parquet or CSV instead."
        )


def _validate_required_columns(filepath: str, required_columns: set[str]) -> None:
    """
    Validates that the file at `filepath` contains all columns in `required_columns`.

    Parameters
    ----------
    filepath : str
        The path to the file to validate.
    required_columns : set[str]
        The set of required column names.

    Raises
    ------
    NotImplementedError
        If the file type is not supported.
    LookupError
        If any required columns are missing.
    """
    extension = Path(filepath).suffix
    if extension == ".parquet":
        output_columns = set(pq.ParquetFile(filepath).schema.names)
    elif extension == ".csv":
        output_columns = set(pd.read_csv(filepath, nrows=5).columns)
    else:
        raise NotImplementedError(
            f"Data file type {extension} is not supported. Convert to Parquet or CSV instead"
        )

    missing_columns = required_columns - output_columns
    if missing_columns:
        raise LookupError(
            f"Data file {filepath} is missing required column(s) {missing_columns}"
        )


def _validate_unique_column(df: pd.DataFrame, column_name: str, filepath: str) -> None:
    """Validates that a column in a DataFrame has unique values.

    Parameters
    ----------
    df : pandas.DataFrame
        The DataFrame to validate.
    column_name : str
        The name of the column to check.
    filepath : str
        The path to the file being validated.

    Raises
    ------
    ValueError
        If the column contains duplicate values.
    """
    if not df[column_name].is_unique:
        raise ValueError(
            f"Data file {filepath} contains duplicate values in the '{column_name}' column."
        )


def _validate_unique_column_set(df: pd.DataFrame, columns: set[str], filepath: str) -> None:
    """
    Validates that the combination of columns in `columns` is unique in the DataFrame.

    Parameters
    ----------
    df : pandas.DataFrame
        The DataFrame to validate.
    columns : set[str]
        The set of column names to check for uniqueness as a group.
    filepath : str
        The path to the file being validated.

    Raises
    ------
    ValueError
        If duplicate rows exist for the given columns.
    """
    if len(df[list(columns)].drop_duplicates()) < len(df):
        raise ValueError(
            f"Data file {filepath} contains duplicate rows with the same values for {columns}."
        )


def validate_input_file_dummy(filepath: str) -> None:
    """Validates an input file to a dummy :class:`~easylink.step.Step`.

    The file must contain the columns: "foo", "bar", and "counter".

    Parameters
    ----------
    filepath : str
        The path to the input file.

    Raises
    ------
    LookupError
        If the file is missing required columns.
    """
    _validate_required_columns(filepath, required_columns={"foo", "bar", "counter"})


def validate_input_dataset_or_known_clusters(filepath: str) -> None:
    """
    Validates a dataset or clusters file based on its filename.

    Parameters
    ----------
    filepath : str
        The path to the input file.

    Raises
    ------
    LookupError, ValueError
        If the file fails validation as a dataset or clusters file.
    """
    filepath = Path(filepath)
    if "clusters" in filepath.stem:
        validate_clusters(filepath)
    else:
        validate_dataset(filepath)


def validate_dataset(filepath: str) -> None:
    """Validates a dataset file.

    - Must be in a tabular format and contain a "Record ID" column.
    - The "Record ID" column must have unique integer values.

    Parameters
    ----------
    filepath : str
        The path to the input dataset file.

    Raises
    ------
    LookupError
        If the file is missing the required "Record ID" column.
    ValueError
        If the "Record ID" column is not unique or not integer dtype.
    """
    _validate_required_columns(filepath, {"Record ID"})
    df = _read_file(filepath)
    _validate_unique_column(df, "Record ID", filepath)

    if not is_integer_dtype(df["Record ID"]):
        raise ValueError(
            f"Data file {filepath} contains non-integer values in the 'Record ID' column."
        )


def validate_datasets_directory(filepath: str) -> None:
    """Validates a directory of input dataset files.

    - Each file in the directory must be in a tabular format and contain a "Record ID" column.
    - The "Record ID" column must have unique values.

    Parameters
    ----------
    filepath : str
        The path to the directory containing input dataset files.

    Raises
    ------
    NotADirectoryError
        If the provided path is not a directory.
    LookupError
        If any file is missing the required "Record ID" column.
    ValueError
        If the "Record ID" column is not unique in any file or if a non-file is present.
    """
    input_path = Path(filepath)
    if not input_path.is_dir():
        raise NotADirectoryError(f"The path {filepath} is not a directory.")

    for file in input_path.iterdir():
        if not file.is_file():
            raise ValueError(f"The path {file} is not a file.")
        validate_dataset(file.name)


def validate_clusters(filepath: str) -> None:
    """Validates a file containing cluster information.

    - The file must contain three columns: "Input Record Dataset", "Input Record ID", and "Cluster ID".
    - "Input Record Dataset" and "Input Record ID", considered as a pair, must have unique values.

    Parameters
    ----------
    filepath : str
        The path to the file containing cluster data.

    Raises
    ------
    LookupError
        If the file is missing required columns.
    ValueError
        If the ("Input Record Dataset", "Input Record ID") pair is not unique.
    """
    _validate_required_columns(
        filepath, {"Input Record Dataset", "Input Record ID", "Cluster ID"}
    )
    df = _read_file(filepath)
    _validate_unique_column_set(df, {"Input Record Dataset", "Input Record ID"}, filepath)


def validate_links(filepath: str) -> None:
    """Validates a file containing link information.

    - The file must contain five columns: "Left Record Dataset", "Left Record ID", "Right Record Dataset", "Right Record ID", and "Probability".
    - "Left Record ID" and "Right Record ID" cannot be equal in a row where "Left Record Dataset" also equals "Right Record Dataset".
    - Rows must be unique, ignoring the Probability column.
    - "Left Record Dataset" must be alphabetically before (or equal to) "Right Record Dataset."
    - "Left Record ID" must be less than "Right Record ID" if "Left Record Dataset" equals "Right Record Dataset".
    - "Probability" values must be between 0 and 1 (inclusive).

    Parameters
    ----------
    filepath : str
        The path to the file containing link data.

    Raises
    ------
    LookupError
        If the file is missing required columns.
    ValueError
        If:
        - "Left Record ID" equals "Right Record ID" in any row where datasets match.
        - Duplicate rows exist with the same "Left Record Dataset", "Left Record ID", "Right Record Dataset", and "Right Record ID".
        - "Left Record Dataset" is not alphabetically before or equal to "Right Record Dataset".
        - "Left Record ID" is not less than "Right Record ID" when datasets match.
        - Values in the "Probability" column are not between 0 and 1 (inclusive).
    """
    _validate_required_columns(
        filepath,
        {
            "Left Record Dataset",
            "Left Record ID",
            "Right Record Dataset",
            "Right Record ID",
            "Probability",
        },
    )
    df = _read_file(filepath)

    _validate_pairs(df, filepath)

    if not df["Probability"].between(0, 1).all():
        raise ValueError(
            f"Data file {filepath} contains values in the 'Probability' column that are not between 0 and 1 (inclusive)."
        )


def _validate_pairs(df: pd.DataFrame, filepath: str) -> None:
    """
    Validates pairs in a DataFrame for link or pairs files.

    Parameters
    ----------
    df : pandas.DataFrame
        The DataFrame to validate.
    filepath : str
        The path to the file being validated.

    Raises
    ------
    ValueError
        If any validation rule for pairs is violated.
    """
    if (
        (df["Left Record Dataset"] == df["Right Record Dataset"])
        & (df["Left Record ID"] == df["Right Record ID"])
    ).any():
        raise ValueError(
            f"Data file {filepath} contains rows where 'Left Record ID' is equal to 'Right Record ID' and 'Left Record Dataset' is equal to 'Right Record Dataset'."
        )

    _validate_unique_column_set(
        df,
        {"Left Record Dataset", "Left Record ID", "Right Record Dataset", "Right Record ID"},
        filepath,
    )

    if not all(df["Left Record Dataset"] <= df["Right Record Dataset"]):
        raise ValueError(
            f"Data file {filepath} contains rows where 'Left Record Dataset' is not alphabetically before or equal to 'Right Record Dataset'."
        )

    if not all(
        (df["Left Record ID"] < df["Right Record ID"])
        | (df["Left Record Dataset"] != df["Right Record Dataset"])
    ):
        raise ValueError(
            f"Data file {filepath} contains rows where 'Left Record ID' is not less than 'Right Record ID', though the records are from the same dataset."
        )


def validate_ids_to_remove(filepath: str) -> None:
    """Validates a file containing IDs to remove.

    - The file must contain a single column: "Record ID".
    - "Record ID" must have unique values.

    Parameters
    ----------
    filepath : str
        The path to the file containing IDs to remove.

    Raises
    ------
    LookupError
        If the file is missing the "Record ID" column.
    ValueError
        If the "Record ID" column is not unique.
    """
    _validate_required_columns(filepath, {"Record ID"})
    df = _read_file(filepath)
    _validate_unique_column(df, "Record ID", filepath)


def validate_records(filepath: str) -> None:
    """Validates a file containing records.

    - A file in a tabular format.
    - The file may have any number of columns.
    - Two columns must be called "Input Record Dataset" and "Input Record ID" and they must have unique values as a pair.

    Parameters
    ----------
    filepath : str
        The path to the file containing records.

    Raises
    ------
    LookupError
        If required columns are missing.
    ValueError
        If the ("Input Record Dataset", "Input Record ID") pair is not unique.
    """
    _validate_required_columns(filepath, {"Input Record Dataset", "Input Record ID"})
    df = _read_file(filepath)
    _validate_unique_column_set(df, {"Input Record Dataset", "Input Record ID"}, filepath)


def validate_blocks(filepath: str) -> None:
    """
    Validates a directory containing blocks.

    Each block subdirectory must contain exactly two files: a records file and a pairs file, both in tabular format.

    Validation checks include:
    - The parent directory must exist and be a directory.
    - Each block subdirectory must contain exactly one records file (filename contains "records") and one pairs file (filename contains "pairs").
    - The records file must have columns "Input Record Dataset" and "Input Record ID" with unique pairs.
    - The pairs file must have columns "Left Record Dataset", "Left Record ID", "Right Record Dataset", and "Right Record ID".
    - All values in ("Left Record Dataset", "Left Record ID") and ("Right Record Dataset", "Right Record ID") must exist in the records file.
    - No row in the pairs file may have "Left Record Dataset" == "Right Record Dataset" and "Left Record ID" == "Right Record ID".
    - All rows in the pairs file must be unique with respect to ("Left Record Dataset", "Left Record ID", "Right Record Dataset", "Right Record ID").
    - "Left Record Dataset" must be alphabetically before or equal to "Right Record Dataset".
    - "Left Record ID" must be less than "Right Record ID" if datasets match.
    - No extra files are allowed in block subdirectories.

    Parameters
    ----------
    filepath : str
        Path to the directory containing block subdirectories.

    Raises
    ------
    NotADirectoryError
        If the provided path is not a directory.
    FileNotFoundError
        If a required records or pairs file is missing in any block.
    LookupError
        If required columns are missing in records or pairs files.
    ValueError
        If:
            - ("Input Record Dataset", "Input Record ID") is not unique in the records file.
            - ("Left Record Dataset", "Left Record ID") or ("Right Record Dataset", "Right Record ID") in the pairs file do not exist in the records file.
            - "Left Record Dataset" == "Right Record Dataset" and "Left Record ID" == "Right Record ID" in any row of the pairs file.
            - Duplicate rows exist in the pairs file.
            - "Left Record Dataset" is not alphabetically before or equal to "Right Record Dataset" in any row.
            - "Left Record ID" is not less than "Right Record ID" when datasets match.
            - Extra files are present in a block subdirectory.
    """
    input_path = Path(filepath)

    if not input_path.is_dir():
        raise NotADirectoryError(f"The path {filepath} is not a directory.")

    for block_dir in filter(lambda d: d.is_dir(), input_path.iterdir()):
        files = {file.stem: file for file in block_dir.iterdir() if file.is_file()}
        records_file = next((f for name, f in files.items() if "records" in name), None)
        pairs_file = next((f for name, f in files.items() if "pairs" in name), None)

        if len(files) > 2:
            raise ValueError(f"Extra file(s) found in block directory {block_dir}.")

        if not records_file or not pairs_file:
            raise FileNotFoundError(
                f"Block directory {block_dir} must contain both a records file and a pairs file."
            )

        # Validate records file
        _validate_required_columns(records_file, {"Input Record Dataset", "Input Record ID"})
        records_df = _read_file(records_file)
        _validate_unique_column_set(
            records_df, {"Input Record Dataset", "Input Record ID"}, records_file
        )

        # Validate pairs file
        _validate_required_columns(
            pairs_file,
            {
                "Left Record Dataset",
                "Left Record ID",
                "Right Record Dataset",
                "Right Record ID",
            },
        )
        pairs_df = _read_file(pairs_file)

        # Check that all (dataset, ID) tuples in pairs exist in records
        record_tuples = set(
            records_df[["Input Record Dataset", "Input Record ID"]].itertuples(
                index=False, name=None
            )
        )
        missing_left = (
            set(
                pairs_df[["Left Record Dataset", "Left Record ID"]].itertuples(
                    index=False, name=None
                )
            )
            - record_tuples
        )
        missing_right = (
            set(
                pairs_df[["Right Record Dataset", "Right Record ID"]].itertuples(
                    index=False, name=None
                )
            )
            - record_tuples
        )
        if missing_left or missing_right:
            raise ValueError(
                f"In block {block_dir}, pairs file {pairs_file} contains records not found in records file {records_file}. "
                f"Missing left records: {missing_left}, missing right records: {missing_right}"
            )

        _validate_pairs(pairs_df, pairs_file)


def validate_dir(filepath: str) -> None:
    """
    Validates that the given path is a directory.

    Parameters
    ----------
    filepath : str
        The path to check.

    Raises
    ------
    NotADirectoryError
        If the path is not a directory.
    """
    input_path = Path(filepath)
    if not input_path.is_dir():
        raise NotADirectoryError(f"The path {filepath} is not a directory.")


def validate_dataset_dir(filepath: str) -> None:
    """
    Validates a directory containing a single dataset file.

    Parameters
    ----------
    filepath : str
        The path to the directory.

    Raises
    ------
    NotADirectoryError
        If the path is not a directory.
    ValueError
        If the directory contains more than one file.
    FileNotFoundError
        If the directory does not contain any files.
    """
    input_path = Path(filepath)
    if not input_path.is_dir():
        raise NotADirectoryError(f"The path {filepath} is not a directory.")

    file_paths = [f for f in input_path.iterdir() if not str(f.stem).startswith(".")]
    if len(file_paths) > 1:
        raise ValueError(f"The directory {input_path} contains more than one file.")
    if len(file_paths) == 0:
        raise FileNotFoundError(f"The directory {input_path} does not contain any files.")

    file_path = file_paths[0]
    validate_dataset(file_path)


def dont_validate(filepath: str) -> None:
    """Placeholder function that performs no validation.

    Parameters
    ----------
    filepath : str
        The path to the file (not used).
    """
    pass



# PIPELINE_SCHEMA: output_dir
# STEP_NAME: step_2_for_output_dir_example
# REQUIREMENTS: pandas==2.1.2 pyarrow

import os
import shutil
from pathlib import Path

import pandas as pd

dir_path = Path(os.environ["DUMMY_CONTAINER_MAIN_INPUT_DIR_PATH"])
saved = False

for i, f in enumerate([f for f in dir_path.iterdir() if f.is_file()]):
    if "snakemake" in str(f):
        continue

    if not saved:
        shutil.copy(f, os.environ["DUMMY_CONTAINER_OUTPUT_PATHS"])
        saved = True

    print(pd.read_parquet(f))



"""Unit tests for the various Step classes.

Notes
-----
These unit tests often instantiate a Step object using parameters that would not
actually pass validation in a real-world scenario (i.e. they do not conform to the pipeline
schema). This is intentional because it's easier to flex complexity and edge cases here
rather than try to get full coverage in the e2e tests. It also allows us to test for
future pipeline schema expansion/flexibility in a relative simple manner now.
"""

from collections.abc import Callable
from typing import Any

import pytest
from layered_config_tree import LayeredConfigTree
from pytest_mock import MockerFixture

from easylink.configuration import Config
from easylink.graph_components import (
    EdgeParams,
    ImplementationGraph,
    InputSlot,
    InputSlotMapping,
    OutputSlot,
    OutputSlotMapping,
)
from easylink.pipeline_schema_constants.development import NODES
from easylink.step import (
    ChoiceStep,
    EmbarrassinglyParallelStep,
    HierarchicalStep,
    IOStep,
    LoopStep,
    ParallelStep,
    Step,
)
from easylink.utilities.aggregator_utils import concatenate_datasets
from easylink.utilities.splitter_utils import split_data_by_size
from easylink.utilities.validation_utils import validate_input_file_dummy

STEP_KEYS = {step.name: step for step in NODES}


@pytest.fixture
def basic_step_params() -> dict[str, Any]:
    return {
        "step_name": "step_1",
        "input_slots": [
            InputSlot(
                "step_1_main_input",
                "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                validate_input_file_dummy,
            )
        ],
        "output_slots": [OutputSlot("step_1_main_output")],
    }


def test_implementation_node_name(
    basic_step_params: dict[str, Any], default_config: Config
) -> None:
    step = Step(**basic_step_params)
    step.set_configuration_state(default_config["pipeline"]["steps"][step.name], {}, {})
    node_name = step.implementation_node_name
    assert node_name == "step_1_python_pandas"

    step.set_parent_step(Step(step_name="foo", name="bar"))
    node_name = step.implementation_node_name
    assert node_name == "bar_step_1_step_1_python_pandas"


@pytest.fixture
def io_step_params() -> dict[str, Any]:
    return {
        "step_name": "io",
        "input_slots": [InputSlot("result", None, validate_input_file_dummy)],
        "output_slots": [OutputSlot("file1")],
    }


def test_io_step_slots(io_step_params: dict[str, Any]) -> None:
    step = IOStep(**io_step_params)
    assert step.name == step.step_name == "io"
    assert step.input_slots == {
        "result": InputSlot("result", None, validate_input_file_dummy)
    }
    assert step.output_slots == {"file1": OutputSlot("file1")}


def test_io_implementation_graph(io_step_params: dict[str, Any]) -> None:
    step = IOStep(**io_step_params)
    implementation_graph = _create_implementation_graph(step)
    _check_nodes_and_edges(implementation_graph, expected_nodes=["io"], expected_edges=[])


def test_basic_step_slots(basic_step_params: dict[str, Any]) -> None:
    step = Step(**basic_step_params)
    assert step.name == step.step_name == "step_1"
    assert step.input_slots == {
        "step_1_main_input": InputSlot(
            "step_1_main_input",
            "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            validate_input_file_dummy,
        )
    }
    assert step.output_slots == {"step_1_main_output": OutputSlot("step_1_main_output")}


def test_basic_step_implementation_graph(
    basic_step_params: dict[str, Any], default_config: Config
) -> None:
    step = Step(**basic_step_params)
    step.set_configuration_state(default_config["pipeline"]["steps"][step.name], {}, {})
    implementation_graph = _create_implementation_graph(step)
    _check_nodes_and_edges(
        implementation_graph, expected_nodes=["step_1_python_pandas"], expected_edges=[]
    )


def test_basic_step_missing_names_raises() -> None:
    with pytest.raises(
        ValueError, match="All Steps must contain a step_name, name, or both."
    ):
        Step(step_name=None)


@pytest.fixture
def hierarchical_step_params() -> dict[str, Any]:
    return {
        "step_name": "step_4",
        "input_slots": [
            InputSlot(
                "step_4_main_input",
                "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                validate_input_file_dummy,
            )
        ],
        "output_slots": [OutputSlot("step_4_main_output")],
        "nodes": [
            Step(
                "step_4a",
                input_slots=[
                    InputSlot(
                        "step_4a_main_input",
                        "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                        validate_input_file_dummy,
                    )
                ],
                output_slots=[OutputSlot("step_4a_main_output")],
            ),
            Step(
                "step_4b",
                input_slots=[
                    InputSlot(
                        "step_4b_main_input",
                        "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                        validate_input_file_dummy,
                    )
                ],
                output_slots=[OutputSlot("step_4b_main_output")],
            ),
        ],
        "edges": [
            EdgeParams("step_4a", "step_4b", "step_4a_main_output", "step_4b_main_input")
        ],
        "input_slot_mappings": [
            InputSlotMapping("step_4_main_input", "step_4a", "step_4a_main_input")
        ],
        "output_slot_mappings": [
            OutputSlotMapping("step_4_main_output", "step_4b", "step_4b_main_output")
        ],
    }


def test_hierarchical_step_slots(hierarchical_step_params: dict[str, Any]) -> None:
    step = HierarchicalStep(**hierarchical_step_params)
    assert step.name == "step_4"
    assert step.input_slots == {
        "step_4_main_input": InputSlot(
            "step_4_main_input",
            "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            validate_input_file_dummy,
        )
    }
    assert step.output_slots == {"step_4_main_output": OutputSlot("step_4_main_output")}


def test_hierarchical_step_implementation_graph(
    hierarchical_step_params: dict[str, Any]
) -> None:
    step = HierarchicalStep(**hierarchical_step_params)
    step_config = LayeredConfigTree(
        {"implementation": {"name": "step_4_python_pandas", "configuration": {}}}
    )
    step.set_configuration_state(step_config, {}, {})
    implementation_graph = _create_implementation_graph(step)
    _check_nodes_and_edges(
        implementation_graph, expected_nodes=["step_4_python_pandas"], expected_edges=[]
    )

    # Test implementation_graph for substeps
    step_config = LayeredConfigTree(
        {
            "substeps": {
                "step_4a": {
                    "implementation": {
                        "name": "step_4a_python_pandas",
                        "configuration": {},
                    }
                },
                "step_4b": {
                    "implementation": {
                        "name": "step_4b_python_pandas",
                        "configuration": {},
                    }
                },
            },
        }
    )
    step.set_configuration_state(step_config, {}, {})
    implementation_graph = _create_implementation_graph(step)
    expected_nodes = [
        "step_4a_python_pandas",
        "step_4b_python_pandas",
    ]
    expected_edges = [
        (
            "step_4a_python_pandas",
            "step_4b_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_4b_main_input",
                    "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_4a_main_output"),
                "filepaths": None,
            },
        ),
    ]
    _check_nodes_and_edges(implementation_graph, expected_nodes, expected_edges)


@pytest.fixture
def loop_step_params() -> dict[str, Any]:
    return {
        "template_step": HierarchicalStep(
            "step_3",
            input_slots=[
                InputSlot(
                    name="step_3_main_input",
                    env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validator=validate_input_file_dummy,
                ),
                InputSlot(
                    name="step_3_secondary_input",
                    env_var="DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                    validator=validate_input_file_dummy,
                ),
            ],
            output_slots=[OutputSlot("step_3_main_output")],
            nodes=[
                Step(
                    step_name="step_3a",
                    input_slots=[
                        InputSlot(
                            name="step_3a_main_input",
                            env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                            validator=validate_input_file_dummy,
                        ),
                        InputSlot(
                            name="step_3a_secondary_input",
                            env_var="DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                            validator=validate_input_file_dummy,
                        ),
                    ],
                    output_slots=[OutputSlot("step_3a_main_output")],
                ),
                Step(
                    step_name="step_3b",
                    input_slots=[
                        InputSlot(
                            name="step_3b_main_input",
                            env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                            validator=validate_input_file_dummy,
                        ),
                        InputSlot(
                            name="step_3b_secondary_input",
                            env_var="DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                            validator=validate_input_file_dummy,
                        ),
                    ],
                    output_slots=[OutputSlot("step_3b_main_output")],
                ),
            ],
            edges=[
                EdgeParams(
                    source_node="step_3a",
                    target_node="step_3b",
                    output_slot="step_3a_main_output",
                    input_slot="step_3b_main_input",
                ),
            ],
            input_slot_mappings=[
                InputSlotMapping(
                    parent_slot="step_3_main_input",
                    child_node="step_3a",
                    child_slot="step_3a_main_input",
                ),
                InputSlotMapping(
                    parent_slot="step_3_secondary_input",
                    child_node="step_3a",
                    child_slot="step_3a_secondary_input",
                ),
                InputSlotMapping(
                    parent_slot="step_3_secondary_input",
                    child_node="step_3b",
                    child_slot="step_3b_secondary_input",
                ),
            ],
            output_slot_mappings=[
                OutputSlotMapping(
                    parent_slot="step_3_main_output",
                    child_node="step_3b",
                    child_slot="step_3b_main_output",
                )
            ],
        ),
        "self_edges": [
            EdgeParams(
                source_node="step_3",
                target_node="step_3",
                output_slot="step_3_main_output",
                input_slot="step_3_main_input",
            )
        ],
    }


def test_loop_step_slots(loop_step_params: dict[str, Any]) -> None:
    step = LoopStep(**loop_step_params)
    assert step.name == step.step_name == "step_3"
    assert isinstance(step, LoopStep)
    assert step.input_slots == {
        "step_3_main_input": InputSlot(
            "step_3_main_input",
            "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            validate_input_file_dummy,
        ),
        "step_3_secondary_input": InputSlot(
            "step_3_secondary_input",
            "DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
            validate_input_file_dummy,
        ),
    }
    assert step.output_slots == {"step_3_main_output": OutputSlot("step_3_main_output")}
    assert isinstance(step.template_step, HierarchicalStep)
    assert step.template_step.name == step.name
    assert step.template_step.input_slots == step.input_slots
    assert step.template_step.output_slots == step.output_slots
    assert step.self_edges == [
        EdgeParams(step.name, step.name, "step_3_main_output", "step_3_main_input")
    ]


def test_loop_implementation_graph(
    mocker: MockerFixture, loop_step_params: dict[str, Any], default_config: Config
) -> None:
    mocker.patch("easylink.implementation.Implementation._load_metadata")
    mocker.patch("easylink.implementation.Implementation.validate", return_value=[])
    step = LoopStep(**loop_step_params)
    step.set_configuration_state(default_config["pipeline"]["steps"][step.name], {}, {})
    implementation_graph = _create_implementation_graph(step)
    _check_nodes_and_edges(
        implementation_graph, expected_nodes=["step_3_python_pandas"], expected_edges=[]
    )

    step_config = LayeredConfigTree(
        {
            "iterate": [
                {
                    "implementation": {
                        "name": "step_3_python_pandas",
                        "configuration": {},
                    },
                },
                {
                    "substeps": {
                        "step_3a": {
                            "implementation": {
                                "name": "step_3a_python_pandas",
                                "configuration": {},
                            },
                        },
                        "step_3b": {
                            "implementation": {
                                "name": "step_3b_python_pandas",
                                "configuration": {},
                            },
                        },
                    },
                },
            ],
        },
    )
    step.set_configuration_state(step_config, {}, {})
    implementation_graph = _create_implementation_graph(step)
    expected_nodes = [
        "step_3_loop_1_step_3_python_pandas",
        "step_3_loop_2_step_3a_step_3a_python_pandas",
        "step_3_loop_2_step_3b_step_3b_python_pandas",
    ]
    expected_edges = [
        (
            "step_3_loop_1_step_3_python_pandas",
            "step_3_loop_2_step_3a_step_3a_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_3a_main_input",
                    "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_3_main_output"),
                "filepaths": None,
            },
        ),
        (
            "step_3_loop_2_step_3a_step_3a_python_pandas",
            "step_3_loop_2_step_3b_step_3b_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_3b_main_input",
                    "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_3a_main_output"),
                "filepaths": None,
            },
        ),
    ]
    _check_nodes_and_edges(implementation_graph, expected_nodes, expected_edges)


@pytest.fixture
def parallel_step_params() -> dict[str, Any]:
    return {
        "template_step": HierarchicalStep(
            "step_1",
            input_slots=[
                InputSlot(
                    name="step_1_main_input",
                    env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validator=validate_input_file_dummy,
                ),
                InputSlot(
                    name="step_1_secondary_input",
                    env_var="DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                    validator=validate_input_file_dummy,
                ),
            ],
            output_slots=[OutputSlot("step_1_main_output")],
            nodes=[
                Step(
                    step_name="step_1a",
                    input_slots=[
                        InputSlot(
                            name="step_1a_main_input",
                            env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                            validator=validate_input_file_dummy,
                        ),
                        InputSlot(
                            name="step_1a_secondary_input",
                            env_var="DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                            validator=validate_input_file_dummy,
                        ),
                    ],
                    output_slots=[OutputSlot("step_1a_main_output")],
                ),
                Step(
                    step_name="step_1b",
                    input_slots=[
                        InputSlot(
                            name="step_1b_main_input",
                            env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                            validator=validate_input_file_dummy,
                        ),
                        InputSlot(
                            name="step_1b_secondary_input",
                            env_var="DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                            validator=validate_input_file_dummy,
                        ),
                    ],
                    output_slots=[OutputSlot("step_1b_main_output")],
                ),
            ],
            edges=[
                EdgeParams(
                    source_node="step_1a",
                    target_node="step_1b",
                    output_slot="step_1a_main_output",
                    input_slot="step_1b_main_input",
                ),
            ],
            input_slot_mappings=[
                InputSlotMapping(
                    parent_slot="step_1_main_input",
                    child_node="step_1a",
                    child_slot="step_1a_main_input",
                ),
                InputSlotMapping(
                    parent_slot="step_1_secondary_input",
                    child_node="step_1a",
                    child_slot="step_1a_secondary_input",
                ),
                InputSlotMapping(
                    parent_slot="step_1_secondary_input",
                    child_node="step_1b",
                    child_slot="step_1b_secondary_input",
                ),
            ],
            output_slot_mappings=[
                OutputSlotMapping(
                    parent_slot="step_1_main_output",
                    child_node="step_1b",
                    child_slot="step_1b_main_output",
                ),
            ],
        ),
    }


def test_parallel_step_slots(parallel_step_params: dict[str, Any]) -> None:
    step = ParallelStep(**parallel_step_params)
    assert step.name == "step_1"
    assert step.input_slots == {
        "step_1_main_input": InputSlot(
            "step_1_main_input",
            "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            validate_input_file_dummy,
        ),
        "step_1_secondary_input": InputSlot(
            "step_1_secondary_input",
            "DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
            validate_input_file_dummy,
        ),
    }
    assert step.output_slots == {"step_1_main_output": OutputSlot("step_1_main_output")}


def test_parallel_step_implementation_graph(
    mocker: MockerFixture, parallel_step_params: dict[str, Any]
) -> None:
    mocker.patch("easylink.implementation.Implementation._load_metadata")
    mocker.patch("easylink.implementation.Implementation.validate", return_value=[])
    step = ParallelStep(**parallel_step_params)
    step_config = LayeredConfigTree(
        {
            "parallel": [
                {
                    "substeps": {
                        "step_1a": {
                            "implementation": {
                                "name": "step_1a_python_pandas",
                            },
                        },
                        "step_1b": {
                            "implementation": {
                                "name": "step_1b_python_pandas",
                            },
                        },
                    },
                    "input_data_file": "input_file_1",
                },
                {
                    "substeps": {
                        "step_1a": {
                            "implementation": {
                                "name": "step_1a_python_pandas",
                            },
                        },
                        "step_1b": {
                            "implementation": {
                                "name": "step_1b_python_pandas",
                            },
                        },
                    },
                    "input_data_file": "input_file_2",
                },
                {
                    "substeps": {
                        "step_1a": {
                            "implementation": {
                                "name": "step_1a_python_pandas",
                            },
                        },
                        "step_1b": {
                            "implementation": {
                                "name": "step_1b_python_pandas",
                            },
                        },
                    },
                    "input_data_file": "input_file_3",
                },
            ],
        },
    )
    step.set_configuration_state(step_config, {}, {})
    implementation_graph = _create_implementation_graph(step)
    expected_nodes = {
        "step_1_parallel_split_1_step_1a_step_1a_python_pandas",
        "step_1_parallel_split_1_step_1b_step_1b_python_pandas",
        "step_1_parallel_split_2_step_1a_step_1a_python_pandas",
        "step_1_parallel_split_2_step_1b_step_1b_python_pandas",
        "step_1_parallel_split_3_step_1a_step_1a_python_pandas",
        "step_1_parallel_split_3_step_1b_step_1b_python_pandas",
    }
    expected_edges = [
        (
            "step_1_parallel_split_1_step_1a_step_1a_python_pandas",
            "step_1_parallel_split_1_step_1b_step_1b_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_1b_main_input",
                    "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_1a_main_output"),
                "filepaths": None,
            },
        ),
        (
            "step_1_parallel_split_2_step_1a_step_1a_python_pandas",
            "step_1_parallel_split_2_step_1b_step_1b_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_1b_main_input",
                    "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_1a_main_output"),
                "filepaths": None,
            },
        ),
        (
            "step_1_parallel_split_3_step_1a_step_1a_python_pandas",
            "step_1_parallel_split_3_step_1b_step_1b_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_1b_main_input",
                    "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_1a_main_output"),
                "filepaths": None,
            },
        ),
    ]
    _check_nodes_and_edges(implementation_graph, expected_nodes, expected_edges)


@pytest.mark.parametrize("step_type", ["parallel", "loop"])
def test_templated_implementation_graph_no_multiplicity(
    step_type,
    parallel_step_params: dict[str, Any],
    loop_step_params: dict[str, Any],
    mocker: MockerFixture,
) -> None:
    """Tests that we can handle TemplatedStep but with no multiplicity."""
    mocker.patch("easylink.implementation.Implementation._load_metadata")
    mocker.patch("easylink.implementation.Implementation.validate", return_value=[])
    if step_type == "parallel":
        step = ParallelStep(**parallel_step_params)
    else:  # loop
        step = LoopStep(**loop_step_params)
    step_config = LayeredConfigTree(
        {
            # No multiplicity, i.e. no "parallel" or "loop" key!
            "substeps": {
                f"{step.name}a": {
                    "implementation": {
                        "name": f"{step.name}a_python_pandas",
                    },
                },
                f"{step.name}b": {
                    "implementation": {
                        "name": f"{step.name}b_python_pandas",
                    },
                },
            },
            "input_data_file": "input_file_1",
        },
    )
    step.set_configuration_state(step_config, {}, {})
    implementation_graph = _create_implementation_graph(step)
    expected_nodes = [
        f"{step.name}a_python_pandas",
        f"{step.name}b_python_pandas",
    ]
    expected_edges = [
        (
            f"{step.name}a_python_pandas",
            f"{step.name}b_python_pandas",
            {
                "input_slot": InputSlot(
                    f"{step.name}b_main_input",
                    "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validate_input_file_dummy,
                ),
                "output_slot": OutputSlot(f"{step.name}a_main_output"),
                "filepaths": None,
            },
        ),
    ]
    _check_nodes_and_edges(implementation_graph, expected_nodes, expected_edges)


@pytest.mark.parametrize("step_type", ["loop", "parallel"])
@pytest.mark.parametrize("single_repeat", [True, False])
def test__duplicate_template_step(
    step_type, single_repeat, loop_step_params, parallel_step_params, mocker: MockerFixture
):
    """Test against _duplicate_template_step.

    This is not an exhaustive test due to the complicated nature of testing
    equality of different attributes between two deep copies. For example,
    SomeClass.foo == SomeOtherClass.foo if foo is the same string even though
    they are different objects in memory. This is not the case for methods as
    well as other attribute types.
    """
    mocker.patch("easylink.implementation.Implementation._load_metadata")
    mocker.patch("easylink.implementation.Implementation.validate", return_value=[])
    if step_type == "loop":
        step = LoopStep(**loop_step_params)
        config_key = "iterate"
    else:  # parallel
        step = ParallelStep(**parallel_step_params)
        config_key = "parallel"
    implementation_config = [
        {
            "implementation": {
                "name": f"{step.name}_python_pandas",
            },
        }
    ] * (1 if single_repeat else 2)
    step_config = LayeredConfigTree(
        {
            config_key: implementation_config,
        },
    )
    step.set_configuration_state(step_config, {}, {})
    step.template_step.set_configuration_state(
        LayeredConfigTree(
            {
                "implementation": {
                    "name": f"{step.name}_python_pandas",
                },
            },
        ),
        {},
        {},
    )

    duplicate_template_step = step._duplicate_template_step()

    special_handle_attrs = [
        "_configuration_state",
        "configuration_state",
        "step_graph",
        "nodes",
    ]
    attrs = [
        attr
        for attr in dir(step.template_step)
        if attr not in special_handle_attrs
        # dunders are too complicated to check
        and not attr.startswith("__")
        # methods are bound to each instance
        and not callable(getattr(step.template_step, attr))
    ]
    for attr in attrs:
        assert getattr(step.template_step, attr) == getattr(duplicate_template_step, attr)

    # Handle the special cases. Just check that they are of the same type and yet
    # not equal (implying that they are bound to different instances)
    for attr in special_handle_attrs:
        assert isinstance(
            getattr(step.template_step, attr),
            type(getattr(duplicate_template_step, attr)),
        )
        assert getattr(step.template_step, attr) != getattr(duplicate_template_step, attr)
        if attr == "nodes":
            assert [node.name for node in step.template_step.nodes] == [
                node.name for node in duplicate_template_step.nodes
            ]


@pytest.fixture
def choice_step_params() -> dict[str, Any]:
    return {
        "step_name": "choice_section",
        "input_slots": [
            InputSlot(
                name="choice_section_main_input",
                env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                validator=validate_input_file_dummy,
            ),
            InputSlot(
                name="choice_section_secondary_input",
                env_var="DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                validator=validate_input_file_dummy,
            ),
        ],
        "output_slots": [OutputSlot("choice_section_main_output")],
        "choices": {
            "simple": {
                "step": HierarchicalStep(
                    step_name="step_4",
                    input_slots=[
                        InputSlot(
                            name="step_4_main_input",
                            env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                            validator=validate_input_file_dummy,
                        ),
                        InputSlot(
                            name="step_4_secondary_input",
                            env_var="DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                            validator=validate_input_file_dummy,
                        ),
                    ],
                    output_slots=[OutputSlot("step_4_main_output")],
                    nodes=[
                        Step(
                            step_name="step_4a",
                            input_slots=[
                                InputSlot(
                                    name="step_4a_main_input",
                                    env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                                    validator=validate_input_file_dummy,
                                ),
                                InputSlot(
                                    name="step_4a_secondary_input",
                                    env_var="DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                                    validator=validate_input_file_dummy,
                                ),
                            ],
                            output_slots=[OutputSlot("step_4a_main_output")],
                        ),
                        Step(
                            step_name="step_4b",
                            input_slots=[
                                InputSlot(
                                    name="step_4b_main_input",
                                    env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                                    validator=validate_input_file_dummy,
                                ),
                                InputSlot(
                                    name="step_4b_secondary_input",
                                    env_var="DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                                    validator=validate_input_file_dummy,
                                ),
                            ],
                            output_slots=[OutputSlot("step_4b_main_output")],
                        ),
                    ],
                    edges=[
                        EdgeParams(
                            source_node="step_4a",
                            target_node="step_4b",
                            output_slot="step_4a_main_output",
                            input_slot="step_4b_main_input",
                        ),
                    ],
                    input_slot_mappings=[
                        InputSlotMapping(
                            parent_slot="step_4_main_input",
                            child_node="step_4a",
                            child_slot="step_4a_main_input",
                        ),
                        InputSlotMapping(
                            parent_slot="step_4_secondary_input",
                            child_node="step_4a",
                            child_slot="step_4a_secondary_input",
                        ),
                        InputSlotMapping(
                            parent_slot="step_4_secondary_input",
                            child_node="step_4b",
                            child_slot="step_4b_secondary_input",
                        ),
                    ],
                    output_slot_mappings=[
                        OutputSlotMapping(
                            parent_slot="step_4_main_output",
                            child_node="step_4b",
                            child_slot="step_4b_main_output",
                        ),
                    ],
                ),
                "input_slot_mappings": [
                    InputSlotMapping(
                        parent_slot="choice_section_main_input",
                        child_node="step_4",
                        child_slot="step_4_main_input",
                    ),
                    InputSlotMapping(
                        parent_slot="choice_section_secondary_input",
                        child_node="step_4",
                        child_slot="step_4_secondary_input",
                    ),
                ],
                "output_slot_mappings": [
                    OutputSlotMapping(
                        parent_slot="choice_section_main_output",
                        child_node="step_4",
                        child_slot="step_4_main_output",
                    ),
                ],
            },
            "complex": {
                "step": HierarchicalStep(
                    step_name="step_5_and_6",
                    nodes=[
                        Step(
                            step_name="step_5",
                            input_slots=[
                                InputSlot(
                                    name="step_5_main_input",
                                    env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                                    validator=validate_input_file_dummy,
                                ),
                            ],
                            output_slots=[OutputSlot("step_5_main_output")],
                        ),
                        # Add a more complicated (unsupported) loop step to ensure flexibility
                        LoopStep(
                            template_step=Step(
                                step_name="step_6",
                                input_slots=[
                                    InputSlot(
                                        name="step_6_main_input",
                                        env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                                        validator=validate_input_file_dummy,
                                    ),
                                ],
                                output_slots=[OutputSlot("step_6_main_output")],
                            ),
                            self_edges=[
                                EdgeParams(
                                    source_node="step_6",
                                    target_node="step_6",
                                    output_slot="step_6_main_output",
                                    input_slot="step_6_main_input",
                                )
                            ],
                        ),
                    ],
                    edges=[
                        EdgeParams(
                            source_node="step_5",
                            target_node="step_6",
                            output_slot="step_5_main_output",
                            input_slot="step_6_main_input",
                        ),
                    ],
                ),
                "input_slot_mappings": [
                    InputSlotMapping(
                        parent_slot="choice_section_main_input",
                        child_node="step_5",
                        child_slot="step_5_main_input",
                    ),
                ],
                "output_slot_mappings": [
                    OutputSlotMapping(
                        parent_slot="choice_section_main_output",
                        child_node="step_6",
                        child_slot="step_6_main_output",
                    ),
                ],
            },
        },
    }


def test_choice_step_slots(choice_step_params: dict[str, Any]) -> None:
    step = ChoiceStep(**choice_step_params)
    assert step.name == "choice_section"
    assert step.input_slots == {
        "choice_section_main_input": InputSlot(
            "choice_section_main_input",
            "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            validate_input_file_dummy,
        ),
        "choice_section_secondary_input": InputSlot(
            "choice_section_secondary_input",
            "DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
            validate_input_file_dummy,
        ),
    }
    assert step.output_slots == {
        "choice_section_main_output": OutputSlot("choice_section_main_output")
    }


def test_simple_choice_step_implementation_graph(choice_step_params: dict[str, Any]) -> None:
    step = ChoiceStep(**choice_step_params)

    # Test implementation_graph for single step (no substeps)
    step_config = LayeredConfigTree(
        {
            "type": "simple",
            "step_4": {
                "implementation": {
                    "name": "step_4_python_pandas",
                },
            },
        }
    )
    # Need to validate in order to set the step graph an mappings prior to calling `set_configuration_state`
    step.validate_step(step_config, {}, {})
    step.set_configuration_state(step_config, {}, {})
    implementation_graph = _create_implementation_graph(step)
    _check_nodes_and_edges(
        implementation_graph, expected_nodes=["step_4_python_pandas"], expected_edges=[]
    )

    # Test implementation_graph for a step with substeps
    step_config = LayeredConfigTree(
        {
            "type": "simple",
            "step_4": {
                "substeps": {
                    "step_4a": {
                        "implementation": {
                            "name": "step_4a_python_pandas",
                        },
                    },
                    "step_4b": {
                        "implementation": {
                            "name": "step_4b_r",
                        },
                    },
                },
            },
        }
    )
    step.set_configuration_state(step_config, {}, {})
    implementation_graph = _create_implementation_graph(step)
    expected_nodes = [
        "step_4a_python_pandas",
        "step_4b_r",
    ]
    expected_edges = [
        (
            "step_4a_python_pandas",
            "step_4b_r",
            {
                "input_slot": InputSlot(
                    "step_4b_main_input",
                    "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_4a_main_output"),
                "filepaths": None,
            },
        ),
    ]
    _check_nodes_and_edges(implementation_graph, expected_nodes, expected_edges)


def test_complex_choice_step_implementation_graph(choice_step_params: dict[str, Any]) -> None:
    step = ChoiceStep(**choice_step_params)

    step_config = LayeredConfigTree(
        {
            "type": "complex",
            "step_5_and_6": {
                "substeps": {
                    "step_5": {
                        "implementation": {
                            "name": "step_5_python_pandas",
                        },
                    },
                    "step_6": {
                        "iterate": [
                            {
                                "implementation": {
                                    "name": "step_6_python_pandas",
                                },
                            },
                            {
                                "implementation": {
                                    "name": "step_6_python_pandas",
                                },
                            },
                        ],
                    },
                },
            },
        },
    )
    # Need to validate in order to set the step graph and mappings prior to calling `set_configuration_state`
    step.validate_step(step_config, {}, {})
    step.set_configuration_state(step_config, {}, {})
    implementation_graph = _create_implementation_graph(step)
    expected_nodes = [
        "step_5_python_pandas",
        "step_6_loop_1_step_6_python_pandas",
        "step_6_loop_2_step_6_python_pandas",
    ]
    expected_edges = [
        (
            "step_5_python_pandas",
            "step_6_loop_1_step_6_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_6_main_input",
                    "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_5_main_output"),
                "filepaths": None,
            },
        ),
        (
            "step_6_loop_1_step_6_python_pandas",
            "step_6_loop_2_step_6_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_6_main_input",
                    "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_6_main_output"),
                "filepaths": None,
            },
        ),
    ]
    _check_nodes_and_edges(implementation_graph, expected_nodes, expected_edges)


@pytest.fixture
def embarrassingly_parallel_step_params() -> dict[str, Any]:
    return {
        "step": Step(
            step_name="step_3",
            input_slots=[
                InputSlot(
                    "step_3_main_input",
                    "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validate_input_file_dummy,
                )
            ],
            output_slots=[OutputSlot("step_3_main_output")],
        ),
        "slot_splitter_mapping": {"step_3_main_input": split_data_by_size},
        "slot_aggregator_mapping": {"step_3_main_output": concatenate_datasets},
    }


def test_embarrassingly_parallel_step_slots(
    embarrassingly_parallel_step_params: dict[str, Any]
) -> None:
    step = EmbarrassinglyParallelStep(**embarrassingly_parallel_step_params)
    assert step.name == "step_3"
    assert step.input_slots == {
        "step_3_main_input": InputSlot(
            "step_3_main_input",
            "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            validate_input_file_dummy,
        ),
    }
    assert step.output_slots == {
        "step_3_main_output": OutputSlot("step_3_main_output"),
    }


def test_embarrassingly_parallel_step_implementation_graph(
    embarrassingly_parallel_step_params: dict[str, Any]
) -> None:
    ep_step = EmbarrassinglyParallelStep(**embarrassingly_parallel_step_params)
    step_config = LayeredConfigTree({"implementation": {"name": "step_3_python_pandas"}})
    ep_step.set_configuration_state(step_config, {}, {})
    implementation_graph = _create_implementation_graph(ep_step)
    expected_nodes = [
        "step_3_step_3_main_input_split",
        "step_3_python_pandas",
        "step_3_aggregate",
    ]
    expected_edges = [
        (
            "step_3_step_3_main_input_split",
            "step_3_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_3_main_input",
                    env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validator=validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_3_step_3_main_input_split_main_output"),
                "filepaths": None,
            },
        ),
        (
            "step_3_python_pandas",
            "step_3_aggregate",
            {
                "input_slot": InputSlot(
                    "step_3_aggregate_main_input",
                    env_var=None,
                    validator=None,
                ),
                "output_slot": OutputSlot("step_3_main_output"),
                "filepaths": None,
            },
        ),
    ]
    _check_nodes_and_edges(implementation_graph, expected_nodes, expected_edges)
    _check_embarrassingly_parallel_details(implementation_graph, ep_step)


@pytest.mark.parametrize(
    "input_slots, output_slots, slot_splitter_mapping, slot_aggregator_mapping, expected_error_msg",
    [
        (
            [
                InputSlot(
                    "main_input",
                    "FOO",
                    validate_input_file_dummy,
                ),
            ],
            [OutputSlot("main_output")],
            {},
            {"main_output": concatenate_datasets},
            "does not have any input slots with a splitter method assigned",
        ),
        (
            [
                InputSlot(
                    "main_input",
                    "FOO",
                    validate_input_file_dummy,
                ),
                InputSlot(
                    "secondary_input",
                    "BAR",
                    validate_input_file_dummy,
                ),
            ],
            [OutputSlot("main_output")],
            {"main_input": split_data_by_size, "secondary_input": split_data_by_size},
            {"main_output": concatenate_datasets},
            "has multiple input slots with splitter methods assigned",
        ),
        (
            [
                InputSlot(
                    "main_input",
                    "FOO",
                    validate_input_file_dummy,
                ),
            ],
            [
                OutputSlot("main_output"),
                OutputSlot("secondary_output"),
            ],
            {"main_input": split_data_by_size},
            {"main_output": concatenate_datasets},
            "has output slots without aggregator methods",
        ),
        (
            [
                InputSlot(
                    "main_input",
                    "FOO",
                    validate_input_file_dummy,
                ),
            ],
            [
                OutputSlot("main_output"),
                OutputSlot("secondary_output"),
            ],
            {},
            {"main_output": concatenate_datasets},
            [
                "does not have any input slots with a splitter method assigned",
                "has output slots without aggregator methods",
            ],
        ),
        (
            [
                InputSlot(
                    "main_input",
                    "FOO",
                    validate_input_file_dummy,
                ),
                InputSlot(
                    "secondary_input",
                    "BAR",
                    validate_input_file_dummy,
                ),
            ],
            [
                OutputSlot("main_output"),
                OutputSlot("secondary_output"),
            ],
            {"main_input": split_data_by_size, "secondary_input": split_data_by_size},
            {"main_output": concatenate_datasets},
            [
                "has multiple input slots with splitter methods assigned",
                "has output slots without aggregator methods",
            ],
        ),
    ],
    ids=[
        "no_splitter",
        "multiple_splitters",
        "missing_aggregators",
        "no_splitter_missing_aggregators",
        "multiple_splitters_missing_aggregators",
    ],
)
def test_embarrassingly_parallel_step__validation(
    input_slots: list[InputSlot],
    output_slots: list[OutputSlot],
    slot_splitter_mapping: dict[str, Callable],
    slot_aggregator_mapping: dict[str, Callable],
    expected_error_msg: str | list[str],
):
    step_params = {
        "step": Step(
            step_name="step",
            input_slots=input_slots,
            output_slots=output_slots,
        ),
        "slot_splitter_mapping": slot_splitter_mapping,
        "slot_aggregator_mapping": slot_aggregator_mapping,
    }
    with pytest.raises(ValueError) as error:
        EmbarrassinglyParallelStep(**step_params)

    error_msg = str(error.value)

    expected = (
        [expected_error_msg] if isinstance(expected_error_msg, str) else expected_error_msg
    )
    for msg in expected:
        assert msg in error_msg


@pytest.fixture
def embarrassingly_parallel_hierarchical_step_params() -> dict[str, Any]:
    return {
        "step": HierarchicalStep(
            step_name="steps_1_2_3",
            input_slots=[
                InputSlot(
                    "steps_1_2_3_main_input",
                    "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validate_input_file_dummy,
                ),
                InputSlot(
                    "steps_1_2_3_secondary_input",
                    "DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                    validate_input_file_dummy,
                ),
            ],
            output_slots=[
                OutputSlot("steps_1_2_3_main_output"),
            ],
            nodes=[
                Step(
                    "step_1",
                    input_slots=[
                        InputSlot(
                            "step_1_main_input",
                            "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                            validate_input_file_dummy,
                        ),
                        InputSlot(
                            "step_1_secondary_input",
                            "DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                            validate_input_file_dummy,
                        ),
                    ],
                    output_slots=[
                        OutputSlot("step_1_main_output"),
                    ],
                ),
                Step(
                    "step_2",
                    input_slots=[
                        InputSlot(
                            "step_2_main_input",
                            "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                            validate_input_file_dummy,
                        ),
                        InputSlot(
                            "step_2_secondary_input",
                            "DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                            validate_input_file_dummy,
                        ),
                    ],
                    output_slots=[
                        OutputSlot("step_2_main_output"),
                    ],
                ),
                Step(
                    "step_3",
                    input_slots=[
                        InputSlot(
                            "step_3_main_input",
                            "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                            validate_input_file_dummy,
                        ),
                        InputSlot(
                            "step_3_secondary_input",
                            "DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                            validate_input_file_dummy,
                        ),
                    ],
                    output_slots=[
                        OutputSlot("step_3_main_output"),
                    ],
                ),
            ],
            edges=[
                EdgeParams("step_1", "step_2", "step_1_main_output", "step_2_main_input"),
                EdgeParams("step_2", "step_3", "step_2_main_output", "step_3_main_input"),
            ],
            input_slot_mappings=[
                InputSlotMapping("steps_1_2_3_main_input", "step_1", "step_1_main_input"),
                InputSlotMapping(
                    "steps_1_2_3_secondary_input", "step_1", "step_1_secondary_input"
                ),
            ],
            output_slot_mappings=[
                OutputSlotMapping("steps_1_2_3_main_output", "step_3", "step_3_main_output"),
            ],
        ),
        "slot_splitter_mapping": {"steps_1_2_3_main_input": split_data_by_size},
        "slot_aggregator_mapping": {
            "steps_1_2_3_main_output": concatenate_datasets,
        },
    }


def test_embarrassingly_parallel_hierarchical_step_implementation_graph(
    embarrassingly_parallel_hierarchical_step_params,
) -> None:
    ep_step = EmbarrassinglyParallelStep(**embarrassingly_parallel_hierarchical_step_params)
    step_config = LayeredConfigTree(
        {
            "substeps": {
                "step_1": {"implementation": {"name": "step_1_python_pandas"}},
                "step_2": {"implementation": {"name": "step_2_python_pandas"}},
                "step_3": {"implementation": {"name": "step_3_python_pandas"}},
            },
        }
    )
    ep_step.set_configuration_state(step_config, {}, {})
    implementation_graph = _create_implementation_graph(ep_step)
    expected_nodes = [
        "steps_1_2_3_steps_1_2_3_main_input_split",
        "step_1_python_pandas",
        "step_2_python_pandas",
        "step_3_python_pandas",
        "steps_1_2_3_aggregate",
    ]
    expected_edges = [
        (
            "steps_1_2_3_steps_1_2_3_main_input_split",
            "step_1_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_1_main_input",
                    env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validator=validate_input_file_dummy,
                ),
                "output_slot": OutputSlot(
                    "steps_1_2_3_steps_1_2_3_main_input_split_main_output"
                ),
                "filepaths": None,
            },
        ),
        (
            "step_1_python_pandas",
            "step_2_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_2_main_input",
                    env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validator=validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_1_main_output"),
                "filepaths": None,
            },
        ),
        (
            "step_2_python_pandas",
            "step_3_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_3_main_input",
                    env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validator=validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_2_main_output"),
                "filepaths": None,
            },
        ),
        (
            "step_3_python_pandas",
            "steps_1_2_3_aggregate",
            {
                "input_slot": InputSlot(
                    "steps_1_2_3_aggregate_main_input",
                    env_var=None,
                    validator=None,
                ),
                "output_slot": OutputSlot("step_3_main_output"),
                "filepaths": None,
            },
        ),
    ]
    _check_nodes_and_edges(implementation_graph, expected_nodes, expected_edges)
    _check_embarrassingly_parallel_details(implementation_graph, ep_step)


@pytest.fixture
def embarrassingly_parallel_loop_step_params(
    loop_step_params: dict[str, Any]
) -> dict[str, Any]:
    return {
        "step": LoopStep(**loop_step_params),
        "slot_splitter_mapping": {"step_3_main_input": split_data_by_size},
        "slot_aggregator_mapping": {"step_3_main_output": concatenate_datasets},
    }


def test_embarrassingly_parallel_loop_step_implementation_graph(
    embarrassingly_parallel_loop_step_params,
    mocker: MockerFixture,
) -> None:
    """Tests an embarrassingly parallel LoopStep.

    The LoopsStep consists of three iterations. The first is a HierarchicalStep
    (step_3a and step_3b) and the remaining two are basic Steps (step_3). We thus expect
    the splitter to be applied to the input slot of loop 1, step_3a and the aggregator
    to be applied to the output slot of loop 3, step 3.
    """
    ep_step = EmbarrassinglyParallelStep(**embarrassingly_parallel_loop_step_params)
    step_config = LayeredConfigTree(
        {
            "iterate": [
                {
                    "substeps": {
                        "step_3a": {"implementation": {"name": "step_3a_python_pandas"}},
                        "step_3b": {"implementation": {"name": "step_3b_python_pandas"}},
                    },
                },
                {"implementation": {"name": "step_3_python_pandas"}},
                {"implementation": {"name": "step_3_python_pandas"}},
            ],
        },
    )
    ep_step.set_configuration_state(step_config, {}, {})
    mocker.patch("easylink.implementation.Implementation._load_metadata")
    implementation_graph = _create_implementation_graph(ep_step)
    expected_nodes = [
        "step_3_step_3_main_input_split",
        "step_3_loop_1_step_3a_step_3a_python_pandas",
        "step_3_loop_1_step_3b_step_3b_python_pandas",
        "step_3_loop_2_step_3_python_pandas",
        "step_3_loop_3_step_3_python_pandas",
        "step_3_aggregate",
    ]
    # NOTE: There are no internal edges to the secondary slot (they are all
    # mapped from the outer HierarchicalStep defined in the `loop_step_params`)
    expected_edges = [
        (
            "step_3_step_3_main_input_split",
            "step_3_loop_1_step_3a_step_3a_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_3a_main_input",
                    env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validator=validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_3_step_3_main_input_split_main_output"),
                "filepaths": None,
            },
        ),
        (
            "step_3_loop_1_step_3a_step_3a_python_pandas",
            "step_3_loop_1_step_3b_step_3b_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_3b_main_input",
                    env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validator=validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_3a_main_output"),
                "filepaths": None,
            },
        ),
        (
            "step_3_loop_1_step_3b_step_3b_python_pandas",
            "step_3_loop_2_step_3_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_3_main_input",
                    env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validator=validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_3b_main_output"),
                "filepaths": None,
            },
        ),
        (
            "step_3_loop_2_step_3_python_pandas",
            "step_3_loop_3_step_3_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_3_main_input",
                    env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validator=validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_3_main_output"),
                "filepaths": None,
            },
        ),
        (
            "step_3_loop_3_step_3_python_pandas",
            "step_3_aggregate",
            {
                "input_slot": InputSlot(
                    "step_3_aggregate_main_input",
                    env_var=None,
                    validator=None,
                ),
                "output_slot": OutputSlot("step_3_main_output"),
                "filepaths": None,
            },
        ),
    ]
    _check_nodes_and_edges(implementation_graph, expected_nodes, expected_edges)
    _check_embarrassingly_parallel_details(implementation_graph, ep_step)


@pytest.fixture
def embarrassingly_parallel_parallel_step_params(
    parallel_step_params: dict[str, Any]
) -> dict[str, Any]:
    return {
        "step": ParallelStep(**parallel_step_params),
        "slot_splitter_mapping": {"step_1_main_input": split_data_by_size},
        "slot_aggregator_mapping": {"step_1_main_output": concatenate_datasets},
    }


def test_embarrassingly_parallel_parallel_step_implementation_graph(
    embarrassingly_parallel_parallel_step_params,
    mocker: MockerFixture,
) -> None:
    """Tests an embarrassingly parallel ParallelStep.

    The ParallelStep consists of three copies, each of which are a HierarchicalStep
    consisting of step_1a and step_1b. We thus expect the splitter to be applied
    to each copy's input slot for 1a and the aggregator to be applied to the output
    slots of each copy.
    """
    ep_step = EmbarrassinglyParallelStep(**embarrassingly_parallel_parallel_step_params)
    step_config = LayeredConfigTree(
        {
            "parallel": [
                {
                    "substeps": {
                        "step_1a": {"implementation": {"name": "step_1a_python_pandas"}},
                        "step_1b": {"implementation": {"name": "step_1b_python_pandas"}},
                    },
                },
                {"implementation": {"name": "step_1_python_pandas"}},
                {"implementation": {"name": "step_1_python_pandas"}},
            ],
        },
    )
    ep_step.set_configuration_state(step_config, {}, {})
    mocker.patch("easylink.implementation.Implementation._load_metadata")
    implementation_graph = _create_implementation_graph(ep_step)
    expected_nodes = [
        "step_1_step_1_main_input_split",
        "step_1_parallel_split_1_step_1a_step_1a_python_pandas",
        "step_1_parallel_split_1_step_1b_step_1b_python_pandas",
        "step_1_parallel_split_2_step_1_python_pandas",
        "step_1_parallel_split_3_step_1_python_pandas",
        "step_1_aggregate",
    ]
    # Map the internal edges of the graph
    # NOTE: There are no internal edges to the secondary slot (they are all
    # mapped from the outer HierarchicalStep defined in the `loop_step_params`)
    expected_edges = [
        # SplitterStep -> Step edges
        (
            "step_1_step_1_main_input_split",
            "step_1_parallel_split_1_step_1a_step_1a_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_1a_main_input",
                    env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validator=validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_1_step_1_main_input_split_main_output"),
                "filepaths": None,
            },
        ),
        (
            "step_1_step_1_main_input_split",
            "step_1_parallel_split_2_step_1_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_1_main_input",
                    env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validator=validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_1_step_1_main_input_split_main_output"),
                "filepaths": None,
            },
        ),
        (
            "step_1_step_1_main_input_split",
            "step_1_parallel_split_3_step_1_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_1_main_input",
                    env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validator=validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_1_step_1_main_input_split_main_output"),
                "filepaths": None,
            },
        ),
        # Step -> Step edges
        (
            "step_1_parallel_split_1_step_1a_step_1a_python_pandas",
            "step_1_parallel_split_1_step_1b_step_1b_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_1b_main_input",
                    "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_1a_main_output"),
                "filepaths": None,
            },
        ),
        # Step -> AggregatorStep edges
        (
            "step_1_parallel_split_1_step_1b_step_1b_python_pandas",
            "step_1_aggregate",
            {
                "input_slot": InputSlot(
                    "step_1_aggregate_main_input",
                    env_var=None,
                    validator=None,
                ),
                "output_slot": OutputSlot("step_1b_main_output"),
                "filepaths": None,
            },
        ),
        (
            "step_1_parallel_split_2_step_1_python_pandas",
            "step_1_aggregate",
            {
                "input_slot": InputSlot(
                    "step_1_aggregate_main_input",
                    env_var=None,
                    validator=None,
                ),
                "output_slot": OutputSlot("step_1_main_output"),
                "filepaths": None,
            },
        ),
        (
            "step_1_parallel_split_3_step_1_python_pandas",
            "step_1_aggregate",
            {
                "input_slot": InputSlot(
                    "step_1_aggregate_main_input",
                    env_var=None,
                    validator=None,
                ),
                "output_slot": OutputSlot("step_1_main_output"),
                "filepaths": None,
            },
        ),
    ]
    _check_nodes_and_edges(implementation_graph, expected_nodes, expected_edges)
    _check_embarrassingly_parallel_details(implementation_graph, ep_step)


####################
# Helper functions #
####################


def _create_implementation_graph(step: Step) -> ImplementationGraph:
    implementation_graph = ImplementationGraph()
    step.add_nodes_to_implementation_graph(implementation_graph)
    step.add_edges_to_implementation_graph(implementation_graph)
    return implementation_graph


def _check_nodes_and_edges(
    implementation_graph: ImplementationGraph,
    expected_nodes: list[str],
    expected_edges: list[tuple[str, str, dict[str, InputSlot | OutputSlot | None]]],
) -> None:
    assert set(implementation_graph.nodes) == set(expected_nodes)
    assert len(implementation_graph.edges) == len(expected_edges)
    for edge in expected_edges:
        assert edge in implementation_graph.edges(data=True)


def _check_embarrassingly_parallel_details(
    implementation_graph: ImplementationGraph,
    ep_step: EmbarrassinglyParallelStep,
) -> None:
    nodes = implementation_graph.nodes
    splitter_node_name = list(nodes)[0]
    aggregator_node_name = list(nodes)[-1]
    implemented_node_names = [
        node for node in list(nodes) if node not in [splitter_node_name, aggregator_node_name]
    ]
    # check splitter node has function defined
    assert (
        nodes[splitter_node_name]["implementation"].splitter_func_name
        == list(ep_step.slot_splitter_mapping.values())[0].__name__
    )
    # check aggregator node has aggregator mappings and also points to splitter
    implementation = nodes[aggregator_node_name]["implementation"]
    output_slot_name = list(implementation.output_slots.keys())[0]
    assert (
        implementation.aggregator_func_name
        == ep_step.slot_aggregator_mapping[output_slot_name].__name__
    )
    assert (
        nodes[aggregator_node_name]["implementation"].splitter_node_name == splitter_node_name
    )
    # check the rest are embarrassingly parallel
    for node in implemented_node_names:
        assert nodes[node]["implementation"].is_embarrassingly_parallel



"""
=========
Filepaths
=========

This module contains commonly-used filepaths and directories.

"""

from pathlib import Path

DEV_IMAGES_DIR = "/mnt/team/simulation_science/priv/engineering/er_ecosystem/images"
"""Path to the directory where the development/dummy pipeline images are stored."""
DEFAULT_IMAGES_DIR = Path.home() / ".easylink_images"
"""Default subdirectory for storing downloaded images."""
IMPLEMENTATION_METADATA = Path(__file__).parent.parent / "implementation_metadata.yaml"
"""Path to the implementation metadata file."""
EASYLINK_TEMP = {"local": Path("/tmp/easylink"), "slurm": Path("/tmp")}
"""Paths to the easylink tmp/ directory to get bound to the container's /tmp directory.
When running on slurm, we bind /tmp (rather than /tmp/easylink) to avoid creating 
a subdir with a prolog script"""
SPARK_SNAKEFILE = Path(__file__).parent / "spark.smk"
"""Path to the Snakemake snakefile containing spark-specific rules."""



# STEP_NAME: schema_alignment
# REQUIREMENTS: pandas==2.1.2 pyarrow pyyaml

import logging
import os
from pathlib import Path

import pandas as pd

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(message)s",
    handlers=[logging.StreamHandler()],
)


def load_file(file_path, file_format=None):
    logging.info(f"Loading file {file_path} with format {file_format}")
    if file_format is None:
        file_format = file_path.split(".")[-1]
    if file_format == "parquet":
        return pd.read_parquet(file_path)
    raise ValueError(f"Unknown file format {file_format}")


# LOAD INPUTS and SAVE OUTPUTS

# DATASETS_DIR_PATHS is list of directories
dataset_dirs = os.environ["DATASETS_DIR_PATHS"].split(",")

datasets = {}

for dir in dataset_dirs:
    for root, dirs, files in os.walk(dir):
        for file in files:
            if file.startswith("."):
                continue
            datasets[Path(file).stem] = load_file(os.path.join(root, file))

records = pd.concat(
    [df.assign(**{"Input Record Dataset": dataset}) for dataset, df in datasets.items()],
    ignore_index=True,
    sort=False,
)

records = records.rename(columns={"Record ID": "Input Record ID"})

# DUMMY_CONTAINER_OUTPUT_PATHS is a single filepath
output_path = os.environ["DUMMY_CONTAINER_OUTPUT_PATHS"]
Path(output_path).parent.mkdir(exist_ok=True, parents=True)

logging.info(f"Writing output to {output_path}")
records.to_parquet(output_path)



# mypy: ignore-errors
import subprocess
import sys

import pytest

from easylink.utilities.paths import DEV_IMAGES_DIR
from tests.conftest import SPECIFICATIONS_DIR


@pytest.mark.slow
@pytest.mark.parametrize(
    "pipeline_specification, implementations",
    [
        # Basic
        (
            "common/pipeline.yaml",
            [
                "step_1_python_pandas",
                "step_2_python_pandas",
                "step_3_python_pandas",
                "step_4_python_pandas",
            ],
        ),
        # Expanded pipeline with loops, parallel splits, hierarchy
        (
            "e2e/pipeline_expanded.yaml",
            [
                "step_1_parallel_split_1_step_1_python_pandas",
                "step_1_parallel_split_2_step_1_python_pandas",
                "step_1_parallel_split_3_step_1_python_pandas",
                "step_2_python_pandas",
                "step_3_loop_1_step_3_python_pandas",
                "step_3_loop_2_step_3_python_pandas",
                "step_4a_python_pandas",
                "step_4b_python_pandas",
            ],
        ),
    ],
)
def test_step_types(
    pipeline_specification, implementations, test_specific_results_dir, capsys
):
    """Tests against various permutations of complex step types.

    The goal is to test that EasyLink generates the correct output implementations
    depending on the configuration; i.e. if we have a 'substeps' key in the config,
    we get an implementation for each (or else we get a single implementation).
    """
    input_data = "common/input_data.yaml"
    computing_environment = "common/environment_local.yaml"
    with capsys.disabled():  # disabled so we can monitor job submissions
        print("\n\n*** STARTING RUN ***\n" f"[{pipeline_specification}]\n")

        cmd = (
            "easylink run "
            f"-p {SPECIFICATIONS_DIR / pipeline_specification} "
            f"-i {SPECIFICATIONS_DIR / input_data} "
            f"-e {SPECIFICATIONS_DIR / computing_environment} "
            f"-o {str(test_specific_results_dir)} "
            f"-I {DEV_IMAGES_DIR} "
            "--no-timestamp "
            "--schema development "
        )
        subprocess.run(
            cmd,
            shell=True,
            stdout=sys.stdout,
            stderr=sys.stderr,
            check=True,
        )
        print("\n\n*** END OF RUN ***\n" f"[{pipeline_specification}]\n")

    assert (test_specific_results_dir / "result.parquet").exists()

    # Check that we get directories for particular implementations
    for implementation in implementations:
        assert (test_specific_results_dir / "diagnostics" / implementation).exists()
        # step_3 is embarrassingly parallel so the final results land in _aggregate/
        aggregate_mapping = {
            "step_3_python_pandas": "step_3_aggregate",
            "step_3_loop_1_step_3_python_pandas": "step_3_loop_1_aggregate",
            "step_3_loop_2_step_3_python_pandas": "step_3_loop_2_aggregate",
        }
        subdir = aggregate_mapping.get(implementation, implementation)
        assert (
            test_specific_results_dir / "intermediate" / subdir / "result.parquet"
        ).exists()



# STEP_NAME: pre-processing
# REQUIREMENTS: pandas==2.1.2 pyarrow pyyaml

import logging
import os
from pathlib import Path

import pandas as pd

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(message)s",
    handlers=[logging.StreamHandler()],
)


def load_file(file_path, file_format=None):
    logging.info(f"Loading file {file_path} with format {file_format}")
    if file_format is None:
        file_format = file_path.split(".")[-1]
    if file_format == "parquet":
        return pd.read_parquet(file_path)
    raise ValueError(f"Unknown file format {file_format}")


# LOAD INPUTS and SAVE OUTPUTS

# DATASET_DIR_PATHS is list of directories, each containing one file
dataset_paths = os.environ["DATASET_DIR_PATHS"].split(",")
logging.info(f"{dataset_paths=}")

# for workaround, choose path based on INPUT_DATASET configuration
splitter_choice = os.environ["INPUT_DATASET"]
logging.info(f"splitter_choice={splitter_choice}")
dataset_path = None
for path in dataset_paths:
    path = Path(path)
    # NOTE: We iterate the dir here, but it should only have one non-hidden
    # file in it. We don't validate that here as it is checked in the validator.
    for path_to_check in path.iterdir():
        if path_to_check.stem == splitter_choice:
            dataset_path = str(path_to_check)
            break

if dataset_path is None:
    raise ValueError(f"No dataset matching {splitter_choice} found")

# DUMMY_CONTAINER_OUTPUT_PATHS is a single path to a directory ('dataset')
results_dir = Path(os.environ["DUMMY_CONTAINER_OUTPUT_PATHS"])
results_dir.mkdir(exist_ok=True, parents=True)

output_path = results_dir / Path(dataset_path).name

dataset = load_file(dataset_path)

# NOTE: No actual pre-processing here, we save as-is.

logging.info(f"Writing output for dataset from input {dataset_path} to {output_path}")
dataset.to_parquet(output_path)



# STEP_NAME: removing_records

# REQUIREMENTS: pandas==2.1.2 pyarrow pyyaml

# PIPELINE_SCHEMA: main

import logging
import os
from pathlib import Path

import pandas as pd

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(message)s",
    handlers=[logging.StreamHandler()],
)


def load_file(file_path, file_format=None):
    logging.info(f"Loading file {file_path} with format {file_format}")
    if file_format is None:
        file_format = file_path.split(".")[-1]
    if file_format == "parquet":
        return pd.read_parquet(file_path)
    raise ValueError(f"Unknown file format {file_format}")


# LOAD INPUTS and SAVE OUTPUTS

# INPUT_DATASETS_AND_INPUT_KNOWN_CLUSTERS_FILE_PATHS is list of filepaths which includes
# the known_clusters filepath due to workaround
dataset_paths = os.environ["INPUT_DATASETS_AND_INPUT_KNOWN_CLUSTERS_FILE_PATHS"].split(",")
dataset_paths = [path for path in dataset_paths if "clusters.parquet" not in Path(path).stem]

# for workaround, choose path based on INPUT_DATASET configuration
splitter_choice = os.environ["INPUT_DATASET"]
dataset_path = None
for path in dataset_paths:
    if splitter_choice == Path(path).stem:
        dataset_path = path
        break
if dataset_path is None:
    raise ValueError(f"No dataset matching {splitter_choice} found")

# IDS_TO_REMOVE_FILE_PATH is a single filepath (Cloneable section)
ids_filepath = os.environ["IDS_TO_REMOVE_FILE_PATH"]
# DUMMY_CONTAINER_OUTPUT_PATHS is a single path to a directory ('dataset')
results_dir = Path(os.environ["DUMMY_CONTAINER_OUTPUT_PATHS"])
results_dir.mkdir(exist_ok=True, parents=True)

dataset = load_file(dataset_path)
ids_to_remove = load_file(ids_filepath)

dataset = dataset[~dataset["Record ID"].isin(ids_to_remove)]

output_path = results_dir / Path(dataset_path).name
logging.info(f"Writing output for dataset from input {dataset_path} to {output_path}")
dataset.to_parquet(output_path)



"""
=====================================
Development Pipeline Schema Constants
=====================================

This module contains the parameters required to instantiate the 
:class:`~easylink.pipeline_schema.PipelineSchema` for the so-called "development" 
pipeline, i.e. the pipeline used strictly for development purposes as opposed to
real entity resolution since it relies on dummy steps, data, and containers.

"""

from easylink.graph_components import (
    EdgeParams,
    InputSlot,
    InputSlotMapping,
    OutputSlot,
    OutputSlotMapping,
)
from easylink.step import (
    ChoiceStep,
    EmbarrassinglyParallelStep,
    HierarchicalStep,
    InputStep,
    LoopStep,
    OutputStep,
    ParallelStep,
    Step,
)
from easylink.utilities.aggregator_utils import concatenate_datasets
from easylink.utilities.splitter_utils import split_data_by_size
from easylink.utilities.validation_utils import validate_input_file_dummy

NODES = [
    InputStep(),
    ParallelStep(
        template_step=Step(
            step_name="step_1",
            input_slots=[
                InputSlot(
                    name="step_1_main_input",
                    env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validator=validate_input_file_dummy,
                ),
            ],
            output_slots=[OutputSlot("step_1_main_output")],
        ),
    ),
    Step(
        step_name="step_2",
        input_slots=[
            InputSlot(
                name="step_2_main_input",
                env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                validator=validate_input_file_dummy,
            )
        ],
        output_slots=[OutputSlot("step_2_main_output")],
    ),
    LoopStep(
        template_step=EmbarrassinglyParallelStep(
            step=Step(
                step_name="step_3",
                input_slots=[
                    InputSlot(
                        name="step_3_main_input",
                        env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                        validator=validate_input_file_dummy,
                    ),
                ],
                output_slots=[
                    OutputSlot(
                        name="step_3_main_output",
                    ),
                ],
            ),
            slot_splitter_mapping={"step_3_main_input": split_data_by_size},
            slot_aggregator_mapping={"step_3_main_output": concatenate_datasets},
        ),
        self_edges=[
            EdgeParams(
                source_node="step_3",
                target_node="step_3",
                output_slot="step_3_main_output",
                input_slot="step_3_main_input",
            )
        ],
    ),
    ChoiceStep(
        step_name="choice_section",
        input_slots=[
            InputSlot(
                name="choice_section_main_input",
                env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                validator=validate_input_file_dummy,
            ),
            InputSlot(
                name="choice_section_secondary_input",
                env_var="DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                validator=validate_input_file_dummy,
            ),
        ],
        output_slots=[OutputSlot("choice_section_main_output")],
        choices={
            "simple": {
                "step": HierarchicalStep(
                    step_name="step_4",
                    input_slots=[
                        InputSlot(
                            name="step_4_main_input",
                            env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                            validator=validate_input_file_dummy,
                        ),
                        InputSlot(
                            name="step_4_secondary_input",
                            env_var="DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                            validator=validate_input_file_dummy,
                        ),
                    ],
                    output_slots=[OutputSlot("step_4_main_output")],
                    nodes=[
                        Step(
                            step_name="step_4a",
                            input_slots=[
                                InputSlot(
                                    name="step_4a_main_input",
                                    env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                                    validator=validate_input_file_dummy,
                                ),
                                InputSlot(
                                    name="step_4a_secondary_input",
                                    env_var="DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                                    validator=validate_input_file_dummy,
                                ),
                            ],
                            output_slots=[OutputSlot("step_4a_main_output")],
                        ),
                        Step(
                            step_name="step_4b",
                            input_slots=[
                                InputSlot(
                                    name="step_4b_main_input",
                                    env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                                    validator=validate_input_file_dummy,
                                ),
                                InputSlot(
                                    name="step_4b_secondary_input",
                                    env_var="DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                                    validator=validate_input_file_dummy,
                                ),
                            ],
                            output_slots=[OutputSlot("step_4b_main_output")],
                        ),
                    ],
                    edges=[
                        EdgeParams(
                            source_node="step_4a",
                            target_node="step_4b",
                            output_slot="step_4a_main_output",
                            input_slot="step_4b_main_input",
                        ),
                    ],
                    input_slot_mappings=[
                        InputSlotMapping(
                            parent_slot="step_4_main_input",
                            child_node="step_4a",
                            child_slot="step_4a_main_input",
                        ),
                        InputSlotMapping(
                            parent_slot="step_4_secondary_input",
                            child_node="step_4a",
                            child_slot="step_4a_secondary_input",
                        ),
                        InputSlotMapping(
                            parent_slot="step_4_secondary_input",
                            child_node="step_4b",
                            child_slot="step_4b_secondary_input",
                        ),
                    ],
                    output_slot_mappings=[
                        OutputSlotMapping(
                            parent_slot="step_4_main_output",
                            child_node="step_4b",
                            child_slot="step_4b_main_output",
                        ),
                    ],
                ),
                "input_slot_mappings": [
                    InputSlotMapping(
                        parent_slot="choice_section_main_input",
                        child_node="step_4",
                        child_slot="step_4_main_input",
                    ),
                    InputSlotMapping(
                        parent_slot="choice_section_secondary_input",
                        child_node="step_4",
                        child_slot="step_4_secondary_input",
                    ),
                ],
                "output_slot_mappings": [
                    OutputSlotMapping(
                        parent_slot="choice_section_main_output",
                        child_node="step_4",
                        child_slot="step_4_main_output",
                    ),
                ],
            },
            "complex": {
                "step": HierarchicalStep(
                    step_name="step_5_and_6",
                    input_slots=[
                        InputSlot(
                            name="step_5_and_6_main_input",
                            env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                            validator=validate_input_file_dummy,
                        ),
                    ],
                    output_slots=[OutputSlot("step_5_and_6_main_output")],
                    nodes=[
                        Step(
                            step_name="step_5",
                            input_slots=[
                                InputSlot(
                                    name="step_5_main_input",
                                    env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                                    validator=validate_input_file_dummy,
                                ),
                            ],
                            output_slots=[OutputSlot("step_5_main_output")],
                        ),
                        Step(
                            step_name="step_6",
                            input_slots=[
                                InputSlot(
                                    name="step_6_main_input",
                                    env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                                    validator=validate_input_file_dummy,
                                ),
                            ],
                            output_slots=[OutputSlot("step_6_main_output")],
                        ),
                    ],
                    edges=[
                        EdgeParams(
                            source_node="step_5",
                            target_node="step_6",
                            output_slot="step_5_main_output",
                            input_slot="step_6_main_input",
                        ),
                    ],
                    input_slot_mappings=[
                        InputSlotMapping(
                            parent_slot="step_5_and_6_main_input",
                            child_node="step_5",
                            child_slot="step_5_main_input",
                        ),
                    ],
                    output_slot_mappings=[
                        OutputSlotMapping(
                            parent_slot="step_5_and_6_main_output",
                            child_node="step_6",
                            child_slot="step_6_main_output",
                        ),
                    ],
                ),
                "input_slot_mappings": [
                    InputSlotMapping(
                        parent_slot="choice_section_main_input",
                        child_node="step_5_and_6",
                        child_slot="step_5_and_6_main_input",
                    ),
                ],
                "output_slot_mappings": [
                    OutputSlotMapping(
                        parent_slot="choice_section_main_output",
                        child_node="step_5_and_6",
                        child_slot="step_5_and_6_main_output",
                    ),
                ],
            },
        },
    ),
    OutputStep(
        input_slots=[
            InputSlot(name="result", env_var=None, validator=validate_input_file_dummy)
        ],
    ),
]
EDGES = [
    EdgeParams(
        source_node="input_data",
        target_node="step_1",
        output_slot="all",
        input_slot="step_1_main_input",
    ),
    EdgeParams(
        source_node="step_1",
        target_node="step_2",
        output_slot="step_1_main_output",
        input_slot="step_2_main_input",
    ),
    EdgeParams(
        source_node="step_2",
        target_node="step_3",
        output_slot="step_2_main_output",
        input_slot="step_3_main_input",
    ),
    EdgeParams(
        source_node="step_3",
        target_node="choice_section",
        output_slot="step_3_main_output",
        input_slot="choice_section_main_input",
    ),
    EdgeParams(
        source_node="choice_section",
        target_node="results",
        output_slot="choice_section_main_output",
        input_slot="result",
    ),
    EdgeParams(
        source_node="input_data",
        target_node="choice_section",
        output_slot="all",
        input_slot="choice_section_secondary_input",
    ),
]
SCHEMA_PARAMS = (NODES, EDGES)



from pathlib import Path
from typing import cast

import pandas as pd
import pyarrow.parquet as pq
import pytest

from easylink.runner import main
from easylink.utilities.data_utils import load_yaml
from easylink.utilities.paths import DEV_IMAGES_DIR
from tests.conftest import SPECIFICATIONS_DIR

COMMON_SPECIFICATIONS_DIR = SPECIFICATIONS_DIR / "common"
EP_SPECIFICATIONS_DIR = SPECIFICATIONS_DIR / "integration" / "embarrassingly_parallel"


@pytest.mark.slow
def test_looping_embarrassingly_parallel_step(test_specific_results_dir: Path) -> None:
    pipeline_specification = EP_SPECIFICATIONS_DIR / "pipeline_loop_step.yaml"
    input_data = COMMON_SPECIFICATIONS_DIR / "input_data.yaml"

    _run_pipeline_and_confirm_finished(
        "looping_ep_step", test_specific_results_dir, pipeline_specification, input_data
    )

    intermediate_results_dir = test_specific_results_dir / "intermediate"
    for loop in [1, 2]:
        step_name = f"step_1_loop_{loop}"
        # ensure that the input was split into two
        assert (
            len(
                list(
                    (intermediate_results_dir / f"{step_name}_step_1_main_input_split").rglob(
                        "result.parquet"
                    )
                )
            )
            == 2
        )
        # ensure that each chunk was processed individually
        assert (
            len(
                list(
                    (intermediate_results_dir / f"{step_name}_step_1_python_pandas").rglob(
                        "result.parquet"
                    )
                )
            )
            == 2
        )
        # check for aggregated file
        assert (
            intermediate_results_dir / f"{step_name}_aggregate" / "result.parquet"
        ).exists()


EP_SECTION_MAPPING = {
    "parallel_step": {
        "pipeline_spec": "pipeline_parallel_step.yaml",
        "schema_name": "ep_parallel_step",
        "implementation_names": [
            "step_1_parallel_split_1_step_1_python_pandas",
            "step_1_parallel_split_2_step_1_python_pandas",
        ],
    },
    "loop_step": {
        "pipeline_spec": "pipeline_loop_step.yaml",
        "schema_name": "ep_loop_step",
        "implementation_names": [
            "step_1_loop_1_step_1_python_pandas",
            "step_1_loop_2_step_1_python_pandas",
        ],
    },
    "hierarchical_step": {
        "pipeline_spec": "pipeline_hierarchical_step.yaml",
        "schema_name": "ep_hierarchical_step",
        "implementation_names": ["step_1a_python_pandas", "step_1b_python_pandas"],
    },
}


@pytest.mark.slow
@pytest.mark.parametrize(
    "step_type, expected_counter, num_rows_multiplier",
    [
        (
            "loop_step",
            2,
            1,
        ),
        (
            "parallel_step",
            # We have two mutually exclusive and collectively exhaustive subsets of
            # input data from the EmbarrassinglyParallelStep, each of which
            # is duplicated twice (from the ParallelStep). The step_1 container
            # is then run exactly one time on each of these four datasets.
            1,
            # The embarrassingly parallel splitting shouldn't increase the number of rows
            # in and of itself, but the underlying parallel step does. We have two
            # splits and so expect there to be twice as many rows in the final result.
            2,
        ),
        (
            "hierarchical_step",
            2,
            # The particular schema used here has two input slots, where the main
            # one gets split into two for embarrassingly parallel processing.
            # The total number of rows in the entire set of input data (which gets
            # fed into both the main and the secondary input slots) is ROWS.
            # The first substep (in each branch of the embarrassingly parallel split)
            # gets 0.5xROWS (due to the splitting) in the main input slot plus 1xROWS
            # in the secondary input slot for a total of 1.5xROWS.
            # The second substep then gets the processed 1.5xROWS from the first
            # substep in the main slot plus 1xROWS in the secondary input slot for
            # a total of 2.5xROWS.
            # The two branches are then aggregated for a total multiplier of 5x.
            5,
        ),
    ],
)
def test_embarrassingly_parallel_sections(
    step_type: str,
    expected_counter: int,
    num_rows_multiplier: int,
    test_specific_results_dir: Path,
) -> None:
    # unpack the mapping
    step_mapping = EP_SECTION_MAPPING[step_type]
    pipeline_spec = cast(str, step_mapping["pipeline_spec"])
    schema_name = cast(str, step_mapping["schema_name"])
    implementation_names = cast(list[str], step_mapping["implementation_names"])

    pipeline_specification = EP_SPECIFICATIONS_DIR / pipeline_spec
    input_data = COMMON_SPECIFICATIONS_DIR / "input_data.yaml"

    _run_pipeline_and_confirm_finished(
        schema_name, test_specific_results_dir, pipeline_specification, input_data
    )

    intermediate_results_dir = test_specific_results_dir / "intermediate"
    # ensure that the input was split into two
    assert (
        len(
            list(
                (intermediate_results_dir / "step_1_step_1_main_input_split").rglob(
                    "result.parquet"
                )
            )
        )
        == 2
    )
    for imp_name in implementation_names:
        # ensure that each chunk was processed individually
        assert len(list((intermediate_results_dir / imp_name).rglob("result.parquet"))) == 2
    # check for aggregated file
    aggregated_filepath = intermediate_results_dir / "step_1_aggregate" / "result.parquet"
    assert aggregated_filepath.exists()
    # check that aggregated file was passed along correctly (i.e. made it to the next
    # step, which for this schema is the output step)
    df_aggregated = pd.read_parquet(aggregated_filepath)
    df_final = pd.read_parquet(test_specific_results_dir / "result.parquet")
    assert df_aggregated.equals(df_final)
    # check that the aggregated file is the expected shape
    input_num_rows = 0
    input_filepaths = load_yaml(input_data)
    for filepath in input_filepaths.values():
        input_num_rows += pq.read_metadata(filepath).num_rows
    assert len(df_final) == input_num_rows * num_rows_multiplier
    assert max(df_final["counter"]) == expected_counter
    assert [column for column in df_final.columns if column.startswith("added_column")] == [
        "added_column_0"
    ] + [f"added_column_{i+1}" for i in range(expected_counter)]


####################
# Helper Functions #
####################


def _run_pipeline_and_confirm_finished(
    schema_name: str,
    results_dir: Path,
    pipeline_specification: Path,
    input_data: Path,
    computing_environment: Path | None = None,
) -> None:
    # Run the pipeline. Snakemake exits at the end so we need to catch that here
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        main(
            command="run",
            pipeline_specification=pipeline_specification,
            input_data=input_data,
            computing_environment=computing_environment,
            results_dir=results_dir,
            images_dir=DEV_IMAGES_DIR,
            schema_name=schema_name,
        )
    assert pytest_wrapped_e.value.code == 0

    assert (results_dir / "result.parquet").exists()
    assert (results_dir / pipeline_specification.name).exists()
    assert (results_dir / input_data.name).exists()
    if computing_environment:
        assert (results_dir / computing_environment.name).exists()



# STEP_NAME: blocking_and_filtering
# REQUIREMENTS: pandas pyarrow splink==4.0.7

import os

import pandas as pd

records = pd.read_parquet(os.environ["RECORDS_FILE_PATH"])

# DUMMY_CONTAINER_OUTPUT_PATHS is a single path to a directory ('dataset')
results_dir = os.environ["DUMMY_CONTAINER_OUTPUT_PATHS"]

import splink

blocking_rules = os.environ["BLOCKING_RULES"].split(",")

from splink import Linker, SettingsCreator

# Create the Splink linker in dedupe mode
settings = SettingsCreator(
    link_type="link_and_dedupe",
    blocking_rules_to_generate_predictions=blocking_rules,
    comparisons=[],
)
from splink import DuckDBAPI

grouped = records.rename(columns={"Input Record ID": "unique_id"}).groupby(
    "Input Record Dataset"
)

db_api = DuckDBAPI()
linker = Linker(
    [df for _, df in grouped],
    settings,
    db_api=db_api,
    input_table_aliases=[name for name, _ in grouped],
)

# Copied/adapted from https://github.com/moj-analytical-services/splink/blob/3eb1921eaff6b8471d3ebacd3238eb514f62c844/splink/internals/linker_components/inference.py#L86-L131
from splink.internals.pipeline import CTEPipeline
from splink.internals.vertically_concatenate import compute_df_concat_with_tf

pipeline = CTEPipeline()

# In duckdb, calls to random() in a CTE pipeline cause problems:
# https://gist.github.com/RobinL/d329e7004998503ce91b68479aa41139
df_concat_with_tf = compute_df_concat_with_tf(linker, pipeline)
pipeline = CTEPipeline([df_concat_with_tf])

blocking_input_tablename_l = "__splink__df_concat_with_tf"
blocking_input_tablename_r = "__splink__df_concat_with_tf"

link_type = linker._settings_obj._link_type

# If exploded blocking rules exist, we need to materialise
# the tables of ID pairs
from splink.internals.blocking import materialise_exploded_id_tables

exploding_br_with_id_tables = materialise_exploded_id_tables(
    link_type=link_type,
    blocking_rules=linker._settings_obj._blocking_rules_to_generate_predictions,
    db_api=linker._db_api,
    splink_df_dict=linker._input_tables_dict,
    source_dataset_input_column=linker._settings_obj.column_info_settings.source_dataset_input_column,
    unique_id_input_column=linker._settings_obj.column_info_settings.unique_id_input_column,
)

from splink.internals.blocking import block_using_rules_sqls

sqls = block_using_rules_sqls(
    input_tablename_l=blocking_input_tablename_l,
    input_tablename_r=blocking_input_tablename_r,
    blocking_rules=linker._settings_obj._blocking_rules_to_generate_predictions,
    link_type=link_type,
    source_dataset_input_column=linker._settings_obj.column_info_settings.source_dataset_input_column,
    unique_id_input_column=linker._settings_obj.column_info_settings.unique_id_input_column,
)

pipeline.enqueue_list_of_sqls(sqls)

blocked_pairs = (
    linker._db_api.sql_pipeline_to_splink_dataframe(pipeline)
    .as_pandas_dataframe()
    .drop(columns=["match_key"])
)

blocked_pairs[["Left Record Dataset", "Left Record ID"]] = blocked_pairs.pop(
    "join_key_l"
).str.split("-__-", n=1, expand=True)
blocked_pairs[["Right Record Dataset", "Right Record ID"]] = blocked_pairs.pop(
    "join_key_r"
).str.split("-__-", n=1, expand=True)
blocked_pairs[["Left Record ID", "Right Record ID"]] = blocked_pairs[
    ["Left Record ID", "Right Record ID"]
].astype(int)

# Now ensure correct ordering
wrong_order_dataset = (
    blocked_pairs["Left Record Dataset"] > blocked_pairs["Right Record Dataset"]
)
id_cols = ["Left Record Dataset", "Left Record ID", "Right Record Dataset", "Right Record ID"]
switched_id_cols = [
    "Right Record Dataset",
    "Right Record ID",
    "Left Record Dataset",
    "Left Record ID",
]
blocked_pairs.loc[wrong_order_dataset, id_cols] = blocked_pairs.loc[
    wrong_order_dataset, switched_id_cols
].values

wrong_order_ids = (
    blocked_pairs["Left Record Dataset"] == blocked_pairs["Right Record Dataset"]
) & (blocked_pairs["Left Record ID"] > blocked_pairs["Right Record ID"])
blocked_pairs.loc[wrong_order_ids, id_cols] = blocked_pairs.loc[
    wrong_order_ids, switched_id_cols
].values
blocked_pairs[["Left Record ID", "Right Record ID"]] = blocked_pairs[
    ["Left Record ID", "Right Record ID"]
].astype(int)

print(blocked_pairs)

from pathlib import Path

output_path = Path(results_dir) / "block_0"
output_path.mkdir(exist_ok=True, parents=True)

records.to_parquet(output_path / "records.parquet", index=False)
blocked_pairs.to_parquet(output_path / "pairs.parquet", index=False)



# STEP_NAME: determining_exclusions

# REQUIREMENTS: pandas==2.1.2 pyarrow pyyaml

# PIPELINE_SCHEMA: main

import logging
import os
from pathlib import Path

import pandas as pd

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(message)s",
    handlers=[logging.StreamHandler()],
)


def load_file(file_path, file_format=None):
    logging.info(f"Loading file {file_path} with format {file_format}")
    if file_format is None:
        file_format = file_path.split(".")[-1]
    if file_format == "parquet":
        return pd.read_parquet(file_path)
    raise ValueError(f"Unknown file format {file_format}")


# LOAD INPUTS

# INPUT_DATASETS_AND_INPUT_KNOWN_CLUSTERS_FILE_PATHS is list of filepaths which includes
# the known_clusters filepath due to workaround
dataset_paths = os.environ["INPUT_DATASETS_AND_INPUT_KNOWN_CLUSTERS_FILE_PATHS"].split(",")
dataset_paths = [path for path in dataset_paths if "clusters" not in Path(path).stem]

# for workaround, choose path based on INPUT_DATASET configuration
splitter_choice = os.environ["INPUT_DATASET"]
dataset_path = None
for path in dataset_paths:
    if splitter_choice == Path(path).stem:
        dataset_path = path
        break
if dataset_path is None:
    raise ValueError(f"No dataset matching {splitter_choice} found")

# KNOWN_CLUSTERS_AND_MAYBE_INPUT_DATASETS_FILE_PATHS is a list of file paths.
# There is one item in it that is a file with "clusters" in the filename.
# That's the only item we're interested in here.
# The other items may be there if this is coming from the user's input,
# due to our workaround for only having one slot of user input.
clusters_filepaths = [
    path
    for path in os.environ["KNOWN_CLUSTERS_AND_MAYBE_INPUT_DATASETS_FILE_PATHS"].split(",")
    if "clusters" in Path(path).stem
]
if len(clusters_filepaths) > 1:
    raise ValueError("Multiple known clusters files found")
if len(clusters_filepaths) == 0:
    raise ValueError("No known clusters file found")

clusters_filepath = clusters_filepaths[0]

clusters_df = load_file(clusters_filepath)

if len(clusters_df) > 0:
    raise ValueError(
        "Default implementation of determining_exclusions passed a non-empty set of known clusters"
    )

# don't need to actually load the dataset,
# since we will just save an empty ids_to_remove

# SAVE OUTPUTS

IDS_TO_REMOVE = pd.DataFrame(columns=["Record ID"])

# DUMMY_CONTAINER_OUTPUT_PATHS is a single path to a file (results.parquet)
results_filepath = os.environ["DUMMY_CONTAINER_OUTPUT_PATHS"]

logging.info(f"Writing output for dataset from input {dataset_path} to {results_filepath}")
IDS_TO_REMOVE.to_parquet(results_filepath)



# STEP_NAME: canonicalizing_and_downstream_analysis

# REQUIREMENTS: pandas==2.1.2 pyarrow pyyaml

# PIPELINE_SCHEMA: main

import logging
import os
from itertools import chain, combinations
from pathlib import Path

import pandas as pd

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(message)s",
    handlers=[logging.StreamHandler()],
)


def load_file(file_path, file_format=None):
    logging.info(f"Loading file {file_path} with format {file_format}")
    if file_format is None:
        file_format = file_path.split(".")[-1]
    if file_format == "parquet":
        return pd.read_parquet(file_path)
    raise ValueError(f"Unknown file format {file_format}")


# LOAD INPUTS and SAVE OUTPUTS

# For dummy we will load the clusters and output (only) them as-is

# CLUSTERS_FILE_PATH is a path to a single file
clusters_path = os.environ["CLUSTERS_FILE_PATH"]
# DUMMY_CONTAINER_OUTPUT_PATHS is a path to a single file (results.parquet)
results_filepath = os.environ["DUMMY_CONTAINER_OUTPUT_PATHS"]

clusters_df = load_file(clusters_path)

logging.info(f"Writing output for dataset from input {clusters_path} to {results_filepath}")
clusters_df.to_parquet(results_filepath)



# mypy: ignore-errors
import pytest

from easylink.pipeline_schema import PipelineSchema
from easylink.pipeline_schema_constants import SCHEMA_PARAMS
from easylink.runner import main
from easylink.utilities.paths import DEV_IMAGES_DIR
from tests.conftest import SPECIFICATIONS_DIR


@pytest.mark.slow
def test_missing_results(test_specific_results_dir, mocker, caplog):
    """Test that the pipeline fails when a step is missing output files."""
    nodes, edges = SCHEMA_PARAMS["integration"]
    mocker.patch(
        "easylink.configuration.Config._get_schema",
        return_value=PipelineSchema("integration", nodes=nodes, edges=edges),
    )

    ## Mock implementation script call to wait 1s instead of running something
    mocker.patch(
        "easylink.implementation.Implementation.script_cmd",
        new_callable=mocker.PropertyMock,
        return_value="sleep 1s",
    )
    with pytest.raises(SystemExit) as exit:
        main(
            command="run",
            pipeline_specification=SPECIFICATIONS_DIR / "integration" / "pipeline.yaml",
            input_data=SPECIFICATIONS_DIR / "common/input_data.yaml",
            computing_environment=SPECIFICATIONS_DIR / "common/environment_local.yaml",
            results_dir=test_specific_results_dir,
            images_dir=DEV_IMAGES_DIR,
        )
    assert exit.value.code == 1
    assert "MissingOutputException" in caplog.text


@pytest.mark.slow
@pytest.mark.parametrize("outputs_specified", [True, False])
def test_outputting_a_directory(outputs_specified: bool, test_specific_results_dir, mocker):
    """Test that the pipeline fails when a step is missing output files."""
    nodes, edges = SCHEMA_PARAMS["output_dir"]
    mocker.patch(
        "easylink.configuration.Config._get_schema",
        return_value=PipelineSchema("output_dir", nodes=nodes, edges=edges),
    )

    pipeline_yaml_name = (
        "pipeline_output_dir.yaml"
        if outputs_specified
        else "pipeline_output_dir_default.yaml"
    )
    step_1_output_subdir = (
        "dummy_step_1_for_output_dir_example/output_dir"
        if outputs_specified
        else "dummy_step_1_for_output_dir_example_default"
    )

    with pytest.raises(SystemExit) as exit:
        main(
            command="run",
            pipeline_specification=SPECIFICATIONS_DIR / "integration" / pipeline_yaml_name,
            input_data=SPECIFICATIONS_DIR / "common/input_data_one_file.yaml",
            computing_environment=SPECIFICATIONS_DIR / "common/environment_local.yaml",
            results_dir=test_specific_results_dir,
            images_dir=DEV_IMAGES_DIR,
        )
    assert exit.value.code == 0
    assert (
        len(
            list(
                (test_specific_results_dir / "intermediate" / step_1_output_subdir).rglob(
                    "*.parquet"
                )
            )
        )
        == 3
    )



# PIPELINE_SCHEMA: output_dir
# STEP_NAME: step_1_for_output_dir_example
# REQUIREMENTS: pandas==2.1.2 pyarrow

import os
from pathlib import Path

import pandas as pd

data = pd.read_parquet(os.environ["STEP_1_MAIN_INPUT_FILE_PATHS"])

print(data)

dir_path = Path(os.environ["DUMMY_CONTAINER_OUTPUT_PATHS"])
dir_path.mkdir(parents=True, exist_ok=True)

for i in range(3):
    data.to_parquet(dir_path / f"result_{i}.parquet")



import pytest

from easylink.utilities.data_utils import download_image

TEST_FILE_RECORD_ID = 15611084
TEST_FILE_NAME = "dummy_file"
TEST_FILE_MD5_CHECKSUM = "7f6be7a98d2ce04e8f7b473926c0bab0"


def test_download_image(tmp_path):
    download_image(tmp_path, TEST_FILE_RECORD_ID, TEST_FILE_NAME, TEST_FILE_MD5_CHECKSUM)

    assert (tmp_path / TEST_FILE_NAME).exists()
    with open(tmp_path / TEST_FILE_NAME, "r") as f:
        content = f.read()
    assert content.strip() == "This is a test file; do not delete."


def test_download_image_raises_incorrect_md5_checksum(tmp_path):
    with pytest.raises(ValueError, match="MD5 checksum does not match"):
        download_image(tmp_path, TEST_FILE_RECORD_ID, TEST_FILE_NAME, "invalid_checksum")



# STEP_NAME: updating_clusters

# REQUIREMENTS: pandas==2.1.2 pyarrow pyyaml networkx

# PIPELINE_SCHEMA: main

import logging
import os
from pathlib import Path

import networkx as nx
import pandas as pd

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(message)s",
    handlers=[logging.StreamHandler()],
)


def load_file(file_path, file_format=None):
    logging.info(f"Loading file {file_path} with format {file_format}")
    if file_format is None:
        file_format = file_path.split(".")[-1]
    if file_format == "parquet":
        return pd.read_parquet(file_path)
    raise ValueError(f"Unknown file format {file_format}")


# LOAD INPUTS and SAVE OUTPUTS

# NEW_CLUSTERS_FILE_PATH is a path to a single file
new_clusters_filepath = os.environ["NEW_CLUSTERS_FILE_PATH"]

# KNOWN_CLUSTERS_AND_MAYBE_INPUT_DATASETS_FILE_PATHS is a list of file paths.
# There is one item in it that is a file with "clusters" in the filename.
# That's the only item we're interested in here.
# The other items may be there if this is coming from the user's input,
# due to our workaround for only having one slot of user input.
known_clusters_filepaths = [
    path
    for path in os.environ["KNOWN_CLUSTERS_AND_MAYBE_INPUT_DATASETS_FILE_PATHS"].split(",")
    if "clusters" in Path(path).stem
]
if len(known_clusters_filepaths) > 1:
    raise ValueError("Multiple known clusters files found")
if len(known_clusters_filepaths) == 0:
    raise ValueError("No known clusters file found")

known_clusters_filepath = known_clusters_filepaths[0]
known_clusters_df = load_file(known_clusters_filepath)

# DUMMY_CONTAINER_OUTPUT_PATHS is a path to a single file (clusters.parquet)
results_filepath = os.environ["DUMMY_CONTAINER_OUTPUT_PATHS"]
Path(results_filepath).parent.mkdir(exist_ok=True, parents=True)

new_clusters_df = load_file(new_clusters_filepath)


def merge_clusters(known_clusters_df, new_clusters_df):
    # Combine both dataframes
    combined_df = pd.concat([known_clusters_df, new_clusters_df], ignore_index=True)

    # Drop records with missing cluster IDs
    combined_df = combined_df.dropna(subset=["Cluster ID"])

    # Group by Cluster ID to get connected records
    cluster_groups = combined_df.groupby("Cluster ID")["Input Record ID"].apply(list)

    # Build a graph of all connections implied by cluster IDs
    G = nx.Graph()
    for group in cluster_groups:
        for i in range(len(group)):
            for j in range(i + 1, len(group)):
                G.add_edge(group[i], group[j])

    # Add isolated nodes (records with unique clusters)
    all_ids = set(combined_df["Input Record ID"])
    G.add_nodes_from(all_ids)

    # Compute connected components
    components = list(nx.connected_components(G))

    # Assign new cluster IDs
    merged_data = []
    for cluster_id, records in enumerate(components, start=1):
        for record_id in records:
            merged_data.append((record_id, cluster_id))

    # Build the final DataFrame
    merged_df = pd.DataFrame(merged_data, columns=["Input Record ID", "Cluster ID"])

    return merged_df


output_df = merge_clusters(known_clusters_df, new_clusters_df)

logging.info(
    f"Writing output for dataset from input {new_clusters_filepath} to {results_filepath}"
)
output_df.to_parquet(results_filepath)



# mypy: ignore-errors
import subprocess
import sys
from pathlib import Path

import pandas as pd
import pytest

from easylink.utilities.general_utils import is_on_slurm
from easylink.utilities.paths import DEV_IMAGES_DIR


@pytest.mark.slow
@pytest.mark.skipif(
    not is_on_slurm(),
    reason="Must be on slurm to run this test.",
)
@pytest.mark.parametrize(
    "pipeline_specification, input_data, computing_environment, correct_results_csv",
    [
        # slurm pipeline_splink_dummy.yaml
        (
            "tests/specifications/e2e/pipeline_splink_dummy.yaml",
            "tests/specifications/e2e/input_data_dummy.yaml",
            "tests/specifications/e2e/environment_slurm.yaml",
            "tests/e2e/pipeline_splink_dummy_results.csv",
        ),
        # local pipeline_splink_dummy.yaml
        (
            "tests/specifications/e2e/pipeline_splink_dummy.yaml",
            "tests/specifications/e2e/input_data_dummy.yaml",
            "tests/specifications/common/environment_local.yaml",
            "tests/e2e/pipeline_splink_dummy_results.csv",
        ),
    ],
)
def test_pipeline_splink_dummy(
    pipeline_specification,
    input_data,
    computing_environment,
    correct_results_csv,
    test_specific_results_dir,
    capsys,
):
    """Tests the 'easylink run' command

    Notes
    -----
    We use various print statements in this test because they show up in the
    Jenkins logs.
    """

    with capsys.disabled():  # disabled so we can monitor job submissions
        print(
            "\n\n*** RUNNING TEST ***\n"
            f"[{pipeline_specification}, {input_data}, {computing_environment}]\n"
        )

        cmd = (
            "easylink run "
            f"-p {pipeline_specification} "
            f"-i {input_data} "
            f"-e {computing_environment} "
            f"-o {str(test_specific_results_dir)} "
            f"-I {DEV_IMAGES_DIR} "
            "--no-timestamp "
            "--schema main "
        )
        subprocess.run(
            cmd,
            shell=True,
            stdout=sys.stdout,
            stderr=sys.stderr,
            check=True,
        )
        final_output = test_specific_results_dir / "result.parquet"
        assert final_output.exists()
        # Check that the results file matches the expected value
        results = (
            pd.read_parquet(final_output)
            .sort_values("Input Record ID")
            .reset_index(drop=True)
        )
        correct_results = pd.read_csv(correct_results_csv)
        # Equal, except for inconsequential differences in order
        assert (
            results.sort_values(["Input Record Dataset", "Input Record ID"])
            .reset_index(drop=True)
            .equals(
                correct_results.sort_values(
                    ["Input Record Dataset", "Input Record ID"]
                ).reset_index(drop=True)
            )
        )

        assert (test_specific_results_dir / Path(pipeline_specification).name).exists()
        assert (test_specific_results_dir / Path(input_data).name).exists()
        assert (test_specific_results_dir / Path(computing_environment).name).exists()

        print(
            "\n\n*** END OF TEST ***\n"
            f"[{pipeline_specification}, {input_data}, {computing_environment}]\n"
        )


@pytest.mark.skip(reason="FIXME [SSCI-2312]: This test is sporadically failing")
@pytest.mark.slow
@pytest.mark.skipif(
    not is_on_slurm(),
    reason="Must be on slurm to run this test.",
)
@pytest.mark.parametrize(
    "pipeline_specification, input_data, computing_environment, splink_results_csv",
    [
        # slurm pipeline_with_fastLink.yaml
        (
            "tests/specifications/e2e/pipeline_with_fastLink.yaml",
            "tests/specifications/e2e/input_data_dummy.yaml",
            "tests/specifications/e2e/environment_slurm.yaml",
            "tests/e2e/pipeline_splink_dummy_results.csv",
        ),
        # local pipeline_with_fastLink.yaml
        (
            "tests/specifications/e2e/pipeline_with_fastLink.yaml",
            "tests/specifications/e2e/input_data_dummy.yaml",
            "tests/specifications/common/environment_local.yaml",
            "tests/e2e/pipeline_splink_dummy_results.csv",
        ),
        # slurm pipeline_cascade.yaml
        (
            "tests/specifications/e2e/pipeline_cascade.yaml",
            "tests/specifications/e2e/input_data_dummy.yaml",
            "tests/specifications/e2e/environment_slurm.yaml",
            "tests/e2e/pipeline_splink_dummy_results.csv",
        ),
        # local pipeline_cascade.yaml
        (
            "tests/specifications/e2e/pipeline_cascade.yaml",
            "tests/specifications/e2e/input_data_dummy.yaml",
            "tests/specifications/common/environment_local.yaml",
            "tests/e2e/pipeline_splink_dummy_results.csv",
        ),
    ],
)
def test_pipelines_same_output_relabeled(
    pipeline_specification,
    input_data,
    computing_environment,
    splink_results_csv,
    test_specific_results_dir,
    capsys,
):
    """Tests the 'easylink run' command

    Notes
    -----
    We use various print statements in this test because they show up in the
    Jenkins logs.
    """

    with capsys.disabled():  # disabled so we can monitor job submissions
        print(
            "\n\n*** RUNNING TEST ***\n"
            f"[{pipeline_specification}, {input_data}, {computing_environment}]\n"
        )

        cmd = (
            "easylink run "
            f"-p {pipeline_specification} "
            f"-i {input_data} "
            f"-e {computing_environment} "
            f"-o {str(test_specific_results_dir)} "
            f"-I {DEV_IMAGES_DIR} "
            "--no-timestamp "
            "--schema main "
        )
        subprocess.run(
            cmd,
            shell=True,
            stdout=sys.stdout,
            stderr=sys.stderr,
            check=True,
        )
        final_output = test_specific_results_dir / "result.parquet"
        assert final_output.exists()
        # Check that the results file matches the expected value
        results = (
            pd.read_parquet(final_output)
            .sort_values("Input Record ID")
            .reset_index(drop=True)
        )
        splink_results = pd.read_csv(splink_results_csv)
        # With very simple, exact-match models, splink and fastLink should find the same
        # results -- and the 1:1 links_to_clusters algorithm shouldn't change that, since the
        # only matches to find are cross-file.
        # However, the clusters won't be *labeled* with the same names.
        # This overly-tricky bit of code checks that the actual clusters induced are the same,
        # whether or not they are labeled the same.
        assert frozenset(
            results.groupby("Cluster ID")["Input Record ID"].apply(frozenset)
        ) == frozenset(
            splink_results.groupby("Cluster ID")["Input Record ID"].apply(frozenset)
        )

        assert (test_specific_results_dir / Path(pipeline_specification).name).exists()
        assert (test_specific_results_dir / Path(input_data).name).exists()
        assert (test_specific_results_dir / Path(computing_environment).name).exists()

        print(
            "\n\n*** END OF TEST ***\n"
            f"[{pipeline_specification}, {input_data}, {computing_environment}]\n"
        )



"""
=============================
Main EasyLink Pipeline Schema
=============================
"""

from easylink.graph_components import (
    EdgeParams,
    InputSlot,
    InputSlotMapping,
    OutputSlot,
    OutputSlotMapping,
)
from easylink.step import (
    HierarchicalStep,
    InputStep,
    LoopStep,
    OutputStep,
    ParallelStep,
    Step,
)
from easylink.utilities.validation_utils import (
    dont_validate,
    validate_blocks,
    validate_clusters,
    validate_dataset_dir,
    validate_ids_to_remove,
    validate_input_dataset_or_known_clusters,
    validate_links,
    validate_records,
)

NODES = [
    # NOTE: In our pipeline schema as documented, there are two inputs: input datasets and known clusters
    # However, due to limitations currently in EasyLink, we can't have multiple output slots on the InputStep.
    # Instead we have a single undifferentiated slot and make it the *implementation's* problem to differentiate
    # based on filename.
    InputStep(),
    LoopStep(
        template_step=HierarchicalStep(
            step_name="entity_resolution",
            input_slots=[
                InputSlot(
                    name="input_datasets",
                    env_var="INPUT_DATASETS_AND_INPUT_KNOWN_CLUSTERS_FILE_PATHS",
                    # NOTE: Since this originates from the InputStep, it will be a *list*
                    # of files, and this validator will be called on *each*
                    # TODO: Change this when https://jira.ihme.washington.edu/browse/MIC-6070 is implemented
                    validator=validate_input_dataset_or_known_clusters,
                ),
                InputSlot(
                    name="known_clusters",
                    env_var="KNOWN_CLUSTERS_AND_MAYBE_INPUT_DATASETS_FILE_PATHS",
                    validator=validate_input_dataset_or_known_clusters,
                ),
            ],
            output_slots=[OutputSlot("clusters")],
            nodes=[
                ParallelStep(
                    # NOTE: Splitters/aggregators on the ParallelStep are implicit!
                    template_step=HierarchicalStep(
                        step_name="determining_exclusions_and_removing_records",
                        directly_implemented=False,
                        input_slots=[
                            InputSlot(
                                name="input_datasets",
                                env_var="INPUT_DATASETS_AND_INPUT_KNOWN_CLUSTERS_FILE_PATHS",
                                validator=validate_input_dataset_or_known_clusters,
                            ),
                            InputSlot(
                                name="known_clusters",
                                env_var="KNOWN_CLUSTERS_AND_MAYBE_INPUT_DATASETS_FILE_PATHS",
                                validator=validate_input_dataset_or_known_clusters,
                            ),
                        ],
                        output_slots=[OutputSlot("datasets")],
                        nodes=[
                            Step(
                                step_name="determining_exclusions",
                                input_slots=[
                                    InputSlot(
                                        name="input_datasets",
                                        env_var="INPUT_DATASETS_AND_INPUT_KNOWN_CLUSTERS_FILE_PATHS",
                                        validator=validate_input_dataset_or_known_clusters,
                                    ),
                                    InputSlot(
                                        name="known_clusters",
                                        env_var="KNOWN_CLUSTERS_AND_MAYBE_INPUT_DATASETS_FILE_PATHS",
                                        validator=validate_input_dataset_or_known_clusters,
                                    ),
                                ],
                                output_slots=[OutputSlot("ids_to_remove")],
                            ),
                            Step(
                                step_name="removing_records",
                                input_slots=[
                                    InputSlot(
                                        name="input_datasets",
                                        env_var="INPUT_DATASETS_AND_INPUT_KNOWN_CLUSTERS_FILE_PATHS",
                                        validator=validate_input_dataset_or_known_clusters,
                                    ),
                                    InputSlot(
                                        name="ids_to_remove",
                                        env_var="IDS_TO_REMOVE_FILE_PATH",
                                        validator=validate_ids_to_remove,
                                    ),
                                ],
                                output_slots=[OutputSlot("dataset")],
                            ),
                        ],
                        edges=[
                            EdgeParams(
                                source_node="determining_exclusions",
                                target_node="removing_records",
                                output_slot="ids_to_remove",
                                input_slot="ids_to_remove",
                            )
                        ],
                        input_slot_mappings=[
                            # NOTE: This is the edge that would normally be split,
                            # but it won't be here, because we don't want it to split
                            # the known clusters to be a separate thing!
                            InputSlotMapping(
                                parent_slot="input_datasets",
                                child_node="determining_exclusions",
                                child_slot="input_datasets",
                            ),
                            InputSlotMapping(
                                parent_slot="known_clusters",
                                child_node="determining_exclusions",
                                child_slot="known_clusters",
                            ),
                            InputSlotMapping(
                                parent_slot="input_datasets",
                                child_node="removing_records",
                                child_slot="input_datasets",
                            ),
                        ],
                        output_slot_mappings=[
                            OutputSlotMapping(
                                # Becomes multiple, after implicit cloneable aggregator
                                parent_slot="datasets",
                                child_node="removing_records",
                                child_slot="dataset",
                            )
                        ],
                    )
                ),
                HierarchicalStep(
                    step_name="clustering",
                    input_slots=[
                        InputSlot(
                            name="datasets",
                            env_var="DATASETS_FILE_PATHS",
                            validator=validate_dataset_dir,
                        ),
                        InputSlot(
                            name="known_clusters",
                            env_var="KNOWN_CLUSTERS_AND_MAYBE_INPUT_DATASETS_FILE_PATHS",
                            validator=validate_input_dataset_or_known_clusters,
                        ),
                    ],
                    output_slots=[OutputSlot("new_clusters")],
                    nodes=[
                        Step(
                            step_name="clusters_to_links",
                            input_slots=[
                                InputSlot(
                                    name="known_clusters",
                                    env_var="KNOWN_CLUSTERS_AND_MAYBE_INPUT_DATASETS_FILE_PATHS",
                                    validator=validate_input_dataset_or_known_clusters,
                                ),
                            ],
                            output_slots=[OutputSlot("known_links")],
                        ),
                        LoopStep(
                            template_step=HierarchicalStep(
                                step_name="linking",
                                input_slots=[
                                    InputSlot(
                                        name="datasets",
                                        env_var="DATASETS_FILE_PATHS",
                                        validator=validate_dataset_dir,
                                    ),
                                    InputSlot(
                                        name="known_links",
                                        env_var="KNOWN_LINKS_FILE_PATH",
                                        validator=validate_links,
                                    ),
                                ],
                                output_slots=[OutputSlot("links")],
                                nodes=[
                                    ParallelStep(
                                        template_step=LoopStep(
                                            template_step=Step(
                                                step_name="pre-processing",
                                                input_slots=[
                                                    InputSlot(
                                                        # NOTE: No splitter here, because
                                                        # not supported by EasyLink;
                                                        # the implementation must do the splitting itself.
                                                        name="dataset",
                                                        env_var="DATASET_DIR_PATHS",
                                                        validator=validate_dataset_dir,
                                                    ),
                                                ],
                                                output_slots=[OutputSlot("dataset")],
                                            ),
                                            self_edges=[
                                                EdgeParams(
                                                    source_node="pre-processing",
                                                    target_node="pre-processing",
                                                    output_slot="dataset",
                                                    input_slot="dataset",
                                                ),
                                            ],
                                        )
                                    ),
                                    Step(
                                        step_name="schema_alignment",
                                        input_slots=[
                                            InputSlot(
                                                name="datasets",
                                                env_var="DATASETS_DIR_PATHS",
                                                validator=validate_dataset_dir,
                                            ),
                                        ],
                                        output_slots=[OutputSlot("records")],
                                    ),
                                    Step(
                                        step_name="blocking_and_filtering",
                                        input_slots=[
                                            InputSlot(
                                                name="records",
                                                env_var="RECORDS_FILE_PATH",
                                                validator=validate_records,
                                            ),
                                            InputSlot(
                                                name="known_links",
                                                env_var="KNOWN_LINKS_FILE_PATH",
                                                validator=validate_links,
                                            ),
                                        ],
                                        output_slots=[OutputSlot("blocks")],
                                    ),
                                    Step(
                                        step_name="evaluating_pairs",
                                        input_slots=[
                                            InputSlot(
                                                name="blocks",
                                                env_var="BLOCKS_DIR_PATH",
                                                validator=validate_blocks,
                                            ),
                                            InputSlot(
                                                name="known_links",
                                                env_var="KNOWN_LINKS_FILE_PATH",
                                                validator=validate_links,
                                            ),
                                        ],
                                        output_slots=[OutputSlot("links")],
                                    ),
                                ],
                                edges=[
                                    EdgeParams(
                                        source_node="pre-processing",
                                        target_node="schema_alignment",
                                        output_slot="dataset",
                                        # NOTE: The implicit ParallelStep aggregator has
                                        # made this multiple (a list)
                                        input_slot="datasets",
                                    ),
                                    EdgeParams(
                                        source_node="schema_alignment",
                                        target_node="blocking_and_filtering",
                                        output_slot="records",
                                        input_slot="records",
                                    ),
                                    EdgeParams(
                                        source_node="blocking_and_filtering",
                                        target_node="evaluating_pairs",
                                        output_slot="blocks",
                                        input_slot="blocks",
                                    ),
                                ],
                                input_slot_mappings=[
                                    InputSlotMapping(
                                        parent_slot="datasets",
                                        child_node="pre-processing",
                                        child_slot="dataset",
                                    ),
                                    InputSlotMapping(
                                        parent_slot="known_links",
                                        child_node="blocking_and_filtering",
                                        child_slot="known_links",
                                    ),
                                    InputSlotMapping(
                                        parent_slot="known_links",
                                        child_node="evaluating_pairs",
                                        child_slot="known_links",
                                    ),
                                ],
                                output_slot_mappings=[
                                    OutputSlotMapping(
                                        parent_slot="links",
                                        child_node="evaluating_pairs",
                                        child_slot="links",
                                    )
                                ],
                            ),
                            self_edges=[
                                EdgeParams(
                                    source_node="linking",
                                    target_node="linking",
                                    output_slot="links",
                                    input_slot="known_links",
                                )
                            ],
                        ),
                        Step(
                            step_name="links_to_clusters",
                            input_slots=[
                                InputSlot(
                                    name="links",
                                    env_var="LINKS_FILE_PATH",
                                    validator=validate_links,
                                ),
                            ],
                            output_slots=[OutputSlot("clusters")],
                        ),
                    ],
                    edges=[
                        EdgeParams(
                            source_node="clusters_to_links",
                            target_node="linking",
                            output_slot="known_links",
                            input_slot="known_links",
                        ),
                        EdgeParams(
                            source_node="linking",
                            target_node="links_to_clusters",
                            output_slot="links",
                            input_slot="links",
                        ),
                    ],
                    input_slot_mappings=[
                        InputSlotMapping(
                            parent_slot="datasets",
                            child_node="linking",
                            child_slot="datasets",
                        ),
                        InputSlotMapping(
                            parent_slot="known_clusters",
                            child_node="clusters_to_links",
                            child_slot="known_clusters",
                        ),
                    ],
                    output_slot_mappings=[
                        OutputSlotMapping(
                            parent_slot="new_clusters",
                            child_node="links_to_clusters",
                            child_slot="clusters",
                        ),
                    ],
                ),
                Step(
                    step_name="updating_clusters",
                    input_slots=[
                        InputSlot(
                            name="new_clusters",
                            env_var="NEW_CLUSTERS_FILE_PATH",
                            validator=validate_clusters,
                        ),
                        InputSlot(
                            name="known_clusters",
                            env_var="KNOWN_CLUSTERS_AND_MAYBE_INPUT_DATASETS_FILE_PATHS",
                            validator=validate_input_dataset_or_known_clusters,
                        ),
                    ],
                    output_slots=[OutputSlot("clusters")],
                ),
            ],
            edges=[
                EdgeParams(
                    source_node="determining_exclusions_and_removing_records",
                    target_node="clustering",
                    output_slot="datasets",
                    input_slot="datasets",
                ),
                EdgeParams(
                    source_node="clustering",
                    target_node="updating_clusters",
                    output_slot="new_clusters",
                    input_slot="new_clusters",
                ),
            ],
            input_slot_mappings=[
                InputSlotMapping(
                    parent_slot="input_datasets",
                    child_node="determining_exclusions_and_removing_records",
                    child_slot="input_datasets",
                ),
                InputSlotMapping(
                    parent_slot="known_clusters",
                    child_node="determining_exclusions_and_removing_records",
                    child_slot="known_clusters",
                ),
                InputSlotMapping(
                    parent_slot="known_clusters",
                    child_node="clustering",
                    child_slot="known_clusters",
                ),
                InputSlotMapping(
                    parent_slot="known_clusters",
                    child_node="updating_clusters",
                    child_slot="known_clusters",
                ),
            ],
            output_slot_mappings=[
                OutputSlotMapping(
                    child_node="updating_clusters",
                    child_slot="clusters",
                    parent_slot="clusters",
                ),
            ],
        ),
        self_edges=[
            EdgeParams(
                source_node="entity_resolution",
                target_node="entity_resolution",
                output_slot="clusters",
                input_slot="known_clusters",
            )
        ],
    ),
    Step(
        step_name="canonicalizing_and_downstream_analysis",
        input_slots=[
            InputSlot(
                name="input_datasets",
                env_var="INPUT_DATASETS_AND_INPUT_KNOWN_CLUSTERS_FILE_PATHS",
                validator=validate_input_dataset_or_known_clusters,
            ),
            InputSlot(
                name="clusters",
                env_var="CLUSTERS_FILE_PATH",
                validator=validate_clusters,
            ),
        ],
        output_slots=[OutputSlot("analysis_output")],
    ),
    OutputStep(
        input_slots=[
            InputSlot(name="analysis_output", env_var=None, validator=dont_validate)
        ],
    ),
]
EDGES = [
    EdgeParams(
        source_node="input_data",
        target_node="entity_resolution",
        output_slot="all",
        input_slot="input_datasets",
    ),
    EdgeParams(
        source_node="input_data",
        target_node="entity_resolution",
        output_slot="all",
        input_slot="known_clusters",
    ),
    EdgeParams(
        source_node="input_data",
        target_node="canonicalizing_and_downstream_analysis",
        output_slot="all",
        input_slot="input_datasets",
    ),
    EdgeParams(
        source_node="entity_resolution",
        target_node="canonicalizing_and_downstream_analysis",
        output_slot="clusters",
        input_slot="clusters",
    ),
    EdgeParams(
        source_node="canonicalizing_and_downstream_analysis",
        target_node="results",
        output_slot="analysis_output",
        input_slot="analysis_output",
    ),
]
SCHEMA_PARAMS = (NODES, EDGES)
